"""
Tests Websocket integration

"""

import pytest
from unittest.mock import MagicMock, patch

from onto import testing_utils
from onto.view import WsMediator
from .color_fixtures import Color, PaletteViewModel, rainbow_vm, color_refs
from .fixtures import setup_app, CTX
import flask_socketio
import time


@pytest.mark.skip
def test_connect(setup_app, rainbow_vm, color_refs):
    app = setup_app

    class Demo(WsMediator):
        pass

    mediator = Demo(view_model_cls=rainbow_vm,
                    mutation_cls=None,
                    namespace="/palette")

    io = flask_socketio.SocketIO(app=app)

    io.on_namespace(mediator)

    client = io.test_client(app=app, namespace='/palette')

    assert client.is_connected(namespace='/palette') is True

    res = client.emit('subscribe_view_model',
                      {"color_names": "cian+magenta+yellow"},
                      namespace='/palette')

    assert client.get_received(namespace="/palette") == \
           [{'args': 'connected', 'name': 'message', 'namespace': '/palette'},
            {'args': '{}', 'name': 'message', 'namespace': '/palette'},
            {'args': '{}', 'name': 'message', 'namespace': '/palette'},
            {'args': [], 'name': 'subscribed', 'namespace': '/palette'},
            {'args': [{'colors': ['cian', 'magenta', 'yellow'],
                       'rainbowName': 'cian-magenta-yellow'}],
             'name': 'updated',
             'namespace': '/palette'}]

    color_refs[0].update({"name": "cyan"})

    testing_utils._wait(factor=.7)

    assert client.get_received(namespace="/palette") == [{'name': 'updated', 'args': [{'colors': ['cyan', 'magenta', 'yellow'], 'rainbowName': 'cyan-magenta-yellow'}], 'namespace': '/palette'}]




#!/usr/bin/env python3
"""Pytest Kafka fixtures."""
import sys
import contextlib
import shutil
import urllib.request
import tarfile
from pathlib import Path



KAFKA_URL = 'https://archive.apache.org/dist/kafka/2.2.0/kafka_2.12-2.2.0.tgz'
KAFKA_TAR = 'kafka.tgz'
KAFKA_TAR_ROOTDIR = 'kafka_2.12-2.2.0'
KAFKA_DIR = 'kafka'


def set_up_kafka():
    """Clean, download Kafka from an official mirror and untar it."""
    clean_kafka()

    print('* Downloading Kafka', file=sys.stderr)
    urllib.request.urlretrieve(KAFKA_URL, KAFKA_TAR)

    print('* Unpacking Kafka', file=sys.stderr)
    with tarfile.open(KAFKA_TAR, 'r') as f:
        f.extractall()

    print('* Renaming:', KAFKA_TAR_ROOTDIR, '→', KAFKA_DIR, file=sys.stderr)
    Path(KAFKA_TAR_ROOTDIR).rename(KAFKA_DIR)
    Path(KAFKA_TAR).unlink()


def clean_kafka():
    """Clean whatever `set_up_kafka` may create."""
    shutil.rmtree(KAFKA_DIR, ignore_errors=True)
    shutil.rmtree(KAFKA_TAR_ROOTDIR, ignore_errors=True)
    with contextlib.suppress(FileNotFoundError):
        Path(KAFKA_TAR).unlink()

if __name__ == "__main__":

    set_up_kafka()



"""
Equivalent implementation for flink stateful functions example - ridesharing
ref: https://github.com/apache/flink-statefun/blob/master/statefun-examples/statefun-ridesharing-example/statefun-ridesharing-example-functions/src/main/java/org/apache/flink/statefun/examples/ridesharing/FnDriver.java
"""
from onto.models.base import Serializable

"""
Rewrite callback-style code to async-await: 
ref: https://www.coreycleary.me/how-to-rewrite-a-callback-function-in-promise-form-and-async-await-form-in-javascript 
"""

from onto.domain_model import DomainModel
from onto.attrs import attrs


class RideshareBase(DomainModel):
    pass


class Passenger(RideshareBase):

    async def request_ride(self, start_geo_cell, end_geo_cell):
        r = Ride.create()  # TODO: implement create
        await r.passenger_joins(
            passenger=self,
            start_geo_cell=start_geo_cell,
            end_geo_cell=end_geo_cell
        )

    class PassengerMessage(DomainModel):
        passenger = attrs.relation('Passenger')

        class RideFailedMessage(Serializable):
            ride = attrs.relation('Ride')

        ride_failed = attrs.embed(RideFailedMessage).optional

        class DriverHasBeenFoundMessage(Serializable):
            driver = attrs.relation('Driver')
            driver_geo_cell = attrs.relation('GeoCell')

        driver_found = attrs.embed(RideFailedMessage).optional

        class RideHasStarted(Serializable):
            driver = attrs.relation('Driver')

        ride_started = attrs.embed(RideHasStarted).optional

        class RideHasEnded(Serializable):
            pass  # TODO: make sure that empty class works

        ride_ended = attrs.embed(RideHasEnded).optional

    async def ride_failed(self, ride: 'Ride'):
        message = self.PassengerMessage.new(
            passenger=self,
            ride_failed=self.PassengerMessage.RideFailedMessage.new(
                ride=ride
            )
        )
        message.save()

    async def driver_joins_ride(self, driver: 'Driver', driver_geo_cell: 'GeoCell'):
        message = self.PassengerMessage.new(
            passenger=self,
            driver_found=self.PassengerMessage.DriverHasBeenFoundMessage.new(
                driver=driver,
                driver_geo_cell=driver_geo_cell
            )
        )
        message.save()

    async def ride_started(self, driver: 'Driver'):
        message = self.PassengerMessage.new(
            passenger=self,
            ride_started=self.PassengerMessage.RideHasStarted.new(
                driver=driver
            )
        )
        message.save()

    async def ride_ended(self):
        message = self.PassengerMessage.new(
            passenger=self,
            ride_started=self.PassengerMessage.RideHasEnded.new()
        )
        message.save()


class DriverRejectsPickupError(RideshareBase, Exception):
    driver = attrs.relation(dm_cls='Driver')
    ride = attrs.relation(dm_cls='Ride')


class Driver(RideshareBase):
    is_taken: bool = attrs.required
    current_ride = attrs.relation(dm_cls='Ride').optional
    current_location: 'GeoCell' = attrs.relation(dm_cls='GeoCell')

    @is_taken.getter
    def is_taken(self):
        # TODO: make better
        return self.current_ride is not None

    async def pickup_passenger(self, ride: 'Ride', passenger: Passenger,
                passenger_start_cell: 'GeoCell',
                passenger_end_cell: 'GeoCell'):
        if self.is_taken:
            raise DriverRejectsPickupError(driver=self, ride=ride)
        self.current_ride = ride

        # "    // We also need to unregister ourselves from the current geo cell we belong to."
        if geo_cell := self.current_location:
            await geo_cell.leave_cell(driver=self)

        await ride.driver_joins(driver=self, driver_location=self.current_location)

        message = self.DriverMessage.new(
            driver=self,
            pickup_passenger=self.DriverMessage.PickupPassengerMessage.new(
                passenger=passenger,
                start_geo_location=passenger_start_cell,
                end_geo_location=passenger_end_cell
            )
        )
        message.save()

    class DriverMessage(RideshareBase):
        driver = attrs.relation('Driver')

        class PickupPassengerMessage(Serializable):
            ride = attrs.relation('Ride')  # TODO: maybe passenger_id
            start_geo_location = attrs.relation('GeoCell')
            end_geo_location = attrs.relation('GeoCell')

        pickup_passenger = attrs.embed(PickupPassengerMessage)

    async def ride_has_started(self):
        await self.current_ride.ride_started(driver=self, driver_geo_cell=self.current_location)

    async def ride_has_ended(self):
        await self.current_ride.ride_ended()

    async def location_is_updated(self, current_geo_cell: 'GeoCell'):
        # TODO: maybe switch to embed:     final int updated = locationUpdate.getLocationUpdate().getCurrentGeoCell();
        updated = current_geo_cell
        last = self.current_location
        if last is None:
            self.current_location = updated
            await updated.join_cell()
            return
        elif last == updated:
            return
        else:
            self.current_location = updated


class Ride(RideshareBase):
    passenger = attrs.relation(dm_cls=Passenger)
    driver = attrs.relation(dm_cls=Driver)

    async def ride_started(self, driver: Driver, driver_geo_cell: 'GeoCell'):
        await self.passenger.ride_started(driver=driver)

    async def passenger_joins(
            self,
            passenger: Passenger,
            start_geo_cell: 'GeoCell',
            end_geo_cell: 'GeoCell'
    ):
        self.passenger = passenger
        MAX_RETRY = 5
        # Ref: https://stackoverflow.com/a/7663441
        for trial in range(MAX_RETRY):
            if driver := start_geo_cell.get_driver():
                try:
                    await driver.pickup_passenger(
                        ride=self,
                        passenger=passenger,
                        passenger_start_cell=start_geo_cell,
                        passenger_end_cell=end_geo_cell
                    )
                except DriverRejectsPickupError as _:
                    # TODO: NOTE difference from java impl
                    """
                    final int startGeoCell = passenger.get().getStartGeoCell();
                    String cellKey = String.valueOf(startGeoCell);
                    context.send(FnGeoCell.TYPE, cellKey, GetDriver.getDefaultInstance());
                    """
                    continue  # to retry
                else:
                    break
        else:
            await passenger.ride_failed(ride=self)

    async def driver_joins(self, driver, driver_location):
        self.driver = driver
        await self.passenger.driver_joins_ride(driver=driver, driver_geo_cell=driver_location)

    async def ride_ended(self, ):
        await self.passenger.ride_ended()
        self.passenger = None
        self.driver = None


class GeoCell(RideshareBase):

    drivers: list = attrs.list(
        value=attrs.relation(dm_cls=Driver)
    )

    async def get_driver(self) -> Driver:
        if len(self.drivers) != 0:
            next_driver = self.drivers[0]
            return next_driver
        else:
            return None

    async def leave_cell(self, driver: Driver):
        self.drivers.remove(driver)

    async def add_driver(self):
        # TODO: mutated local variable vs mutated instance state;
        #  may cause difference in behavior
        if self.drivers is None:
            self.drivers = list()
        self.drivers.append(Driver)

    join_cell = add_driver






from collections import namedtuple


RelationshipReference = namedtuple(
    "RelationshipReference",
    ['doc_ref', 'nested', 'obj', 'obj_type'],
    defaults=(None, None, None, None)
)

EmbeddedElement = namedtuple(
    "EmbeddedElement",
    ['d', 'obj_cls', 'obj'],
    defaults=(None, None, None)
)



import functools
from abc import abstractmethod

import typing
from aiokafka import AIOKafkaConsumer, AIOKafkaProducer

from . import Mediator
from .. import view_model
import graphql
from gql import gql, subscribe, mutate



# @subscribe
# async def post_added(parent, info, **kwargs):
#     """
#
#     # Write your query or mutation here
#     subscription onPostAdded {
#       postAdded(author: "hello") {
#         author
#       }
#     }
#
#     :param parent:
#     :param info:
#     :param kwargs:
#     :return:
#     """
#     while True:
#         d = await q.get()
#         yield {
#             'postAdded': d
#         }
#
#
# @mutate
# async def add_post(parent, info, **kwargs):
#     """
#
#     # Write your query or mutation here
#     mutation M($author: String, $comment: String) {
#       addPost(author: $author, comment: $comment) {
#         author
#         comment
#       }
#     }
#
#     # Variables
#     {
#       "author": "hello",
#       "comment": "world"
#     }
#
#     :param parent:
#     :param info:
#     :param kwargs:
#     :return:
#     """
#     await q.put(kwargs)
#     return kwargs


async def _kafka_subscribe(topic_name):
    consumer = AIOKafkaConsumer(
        topic_name,
        bootstrap_servers='localhost:9092',
        # group_id="my-group"
    )
    # Get cluster layout and join group `my-group`
    await consumer.start()
    # TODO: make first yield when listener has started
    try:
        # Consume messages
        async for msg in consumer:
            yield msg
            # print("consumed: ", msg.topic, msg.partition, msg.offset,
            #       msg.key, msg.value, msg.timestamp)
    finally:
        # Will leave consumer group; perform autocommit if enabled.
        await consumer.stop()


async def _kafka_publish(topic_name, value):
    producer = AIOKafkaProducer(
        bootstrap_servers='localhost:9092',
        # group_id="my-group"
    )
    await producer.send(topic=topic_name, value=value)


# class TaskManagementMixin:
#
#     # NOTE: queues here are Last-In-First-Out
#     queues: Dict[str, asyncio.LifoQueue] = dict()
#
#     def _enqueue(self, topic_name, item, ):
#         q = self.queues[topic_name]
#         q.put(item=item)
#
#     def _clear_queue(self, q: asyncio.LifoQueue):
#         while not q.empty():
#             q.get_nowait()
#
#     def _dequeue(self, topic_name):
#         """
#         (Fast-forwarding)
#
#         Clears all but the last element in the queue;
#         Returns the last element.
#         :return:
#         """
#         q = self.queues[topic_name]
#         res = q.get()
#         self._clear_queue(q)
#         return res
#


def graphql_ot_from_view_model(t: typing.Type[view_model.ViewModel]) -> graphql.GraphQLObjectType:
    import marshmallow
    schema: marshmallow.Schema = t.get_schema_obj()
    ot = graphql.GraphQLObjectType(
        name=t.__name__,  # TODO: refactor
        fields={
            k: graphql.GraphQLField(graphql.GraphQLString, )
            for k, v in schema.fields.items()
        }
    )
    return ot


# from kafka import KafkaProducer
#
# # assert isinstance(kafka_server, KafkaProducer)
# kafka_server = KafkaProducer(bootstrap_servers='localhost:9092', api_version=(0, 10, 1))


class GraphQLMediator(Mediator):
    """
    Assumptions:
        - only latest state matters for `state = f([state_snapshot])`

    """

    view_model_cls: typing.Type[view_model.ViewModel] = None
    """ Abstract property """

    @classmethod
    @functools.lru_cache(maxsize=None)
    def _get_graphql_field(cls):
        async def add_post(parent, info, **kwargs):
            await _kafka_publish('my-topic', 'world')

        async def post_added(root, info):
            topic_name = 'my-topic'  # TODO: implement
            gen = _kafka_subscribe(topic_name=topic_name)
            async for msg in gen:
                yield cls._invoke(msg=msg)

        # graphql.GraphQLFieldResolver

        return graphql.GraphQLField(
            graphql_ot_from_view_model(cls.view_model_cls),
            resolve=add_post, subscribe=post_added
        )

    @classmethod
    def _get_graphql_schema(cls) -> graphql.GraphQLSchema:

        # schema = graphql.GraphQLObjectType(
        #     subscription=subscribe_content
        # )

        # @mutate
        #
        #
        # @subscribe

        schema = graphql.GraphQLSchema(
            query=graphql.GraphQLObjectType('posts', {
                    "content": cls._get_graphql_field()
                }),
            mutation=graphql.GraphQLObjectType('addPost', {
                    "content": cls._get_graphql_field()
                }),
            subscription=graphql.GraphQLObjectType(
                "postAdded",
                {
                    "content": cls._get_graphql_field()
                },
            ),
        )
        return schema

    @classmethod
    def start(cls):
        """ Registers a GraphQL service
        TODO: catch when user starts the same mediator twice

        :return:
        """

        async def on_startup():
            from asyncio.queues import Queue
            global q
            q = Queue()

        async def shutdown():
            pass

        from stargql import GraphQL
        app = GraphQL(
            schema=cls._get_graphql_schema(),
            on_startup=[on_startup],
            on_shutdown=[shutdown]
        )
        return app

        # app.add_url_rule(rule, view_func=GraphQLView.as_view(
        #     name,
        #     schema=schema,
        #     graphiql=True,
        # ))

    @classmethod
    def _invoke(cls, msg):
        cls.invoke(msg.topic, msg.key, msg.value)

    @classmethod
    @abstractmethod
    def invoke(cls, *args, **kwargs):
        pass


class LivenessMediator(Mediator):

    @classmethod
    def start(cls):
        # TODO: recover cls.src.start()
        # cls.src.start()
        # s = cls.subscribe_user_view.start()
        from onto.sink.graphql import graph_schema
        import graphql
        liveness = graph_schema(
            op_type='Query',
            name='liveness',
            graphql_object_type=graphql.GraphQLObjectType(
                name='Liveness',
                fields={
                    'alive': graphql.GraphQLField(
                        graphql.GraphQLBoolean,
                        resolve=lambda *args, **kwargs: True),
                }
            ),
            args=dict()
        )

        return [liveness]



import pytest


def test_unwrap():
    # NEEDS: pip install phonenumbers
    # Otherwise, this test is skipped.

    import phonenumbers

    from onto.models.base import Serializable

    class PhoneNumber(Serializable):

        class Meta:
            unwrap = 'formatted_number'

        from onto.attrs import attrs
        formatted_number = attrs.string.optional
        number_object: phonenumbers.PhoneNumber = attrs.internal.optional
        with attrs.easy_init(lambda self: None) as attrs:
            country_code: str = attrs.optional
            national_number: str = attrs.optional
            extension: str = attrs.optional
            raw_input: str = attrs.optional

        @formatted_number.getter
        def formatted_number(self):
            return phonenumbers.format_number(
                self.number_object,
                phonenumbers.PhoneNumberFormat.INTERNATIONAL
            )

        @formatted_number.setter
        def formatted_number(self, formatted_number):
            self._attrs.number_object = phonenumbers.parse(
                formatted_number
            )

        @number_object.getter
        def number_object(self):
            if not hasattr(self._attrs, 'number_object'):
                return phonenumbers.PhoneNumber(
                    country_code=self.country_code,
                    national_number=self.national_number,
                    extension=self.extension,
                    raw_input=self.raw_input,
                )
            else:
                return self._attrs.number_object

    number = PhoneNumber.new(
        country_code='+86',
        national_number='13212345678'
    )

    assert number.formatted_number == '+86 132 1234 5678'
    assert number.to_dict() == '+86 132 1234 5678'
    assert PhoneNumber.from_dict('+86 132 1234 5678').formatted_number == '+86 132 1234 5678'



from onto.attrs.unit import Monad, MonadContext


def test_monad():
    r = Monad._get_root()
    str_optional = r.str.optional
    assert str_optional.properties.type_cls == str
    assert not str_optional.properties.import_required
    assert not str_optional.properties.export_required

    s_int_optional = r.str.int.optional
    assert s_int_optional.properties.type_cls == int
    assert not s_int_optional.properties.import_required
    assert not s_int_optional.properties.export_required


def test_monad_inherit():
    r = Monad._get_root()
    int_a = r.optional
    assert not hasattr(int_a.properties, 'type_cls')
    assert not int_a.properties.import_required

    with MonadContext.context().int:
        assert int_a.properties.type_cls == int


def test_monad_list():

    r = Monad._get_root()
    li = r.list(
        value=lambda a: a.str.optional
    ).required
    assert li.properties.import_required
    assert li.properties.export_required

    assert li.properties.list_value.properties.type_cls == str
    assert not li.properties.list_value.properties.import_required
    assert not li.properties.list_value.properties.export_required


def test_list_decorate():

    r = Monad._get_root()
    li2 = r.list(
        value=lambda a: a.optional
    ).required

    assert not li2.properties.list_value.properties.import_required
    assert not li2.properties.list_value.properties.export_required

    assert not hasattr(li2.properties.list_value.properties, 'type_cls')
    with MonadContext.context().data_key('my_list'):
        assert li2.properties.data_key == 'my_list'
        assert not hasattr(li2.properties.list_value.properties, 'type_cls')


def test_list_inherit_complex():
    r = Monad._get_root()
    li2 = r.list(
        value=lambda a: a.optional
    ).required

    assert not li2.properties.list_value.properties.import_required
    assert not li2.properties.list_value.properties.export_required

    assert not hasattr(li2.properties.list_value.properties, 'type_cls')
    with MonadContext.context().list(
        value=lambda a: a.str
    ):
        assert li2.properties.list_value.properties.type_cls == str



from examples.meeting_room.domain_models import Ticket, User
from examples.meeting_room.domain_models.location import Location
from examples.meeting_room.domain_models.meeting import Meeting
from onto import view_model, attrs
from onto.view.mutation import PatchMutation
from onto.store import Store, reference


class MeetingSessionStore(Store):
    tickets = reference(dm_cls=Ticket, many=True)
    users = reference(dm_cls=User, many=True)
    meeting = reference(dm_cls=Meeting)
    location = reference(dm_cls=Location)


class MeetingSessionMixin:

    class Meta:
        exclude = ('obj_type',)

    latitude = attrs.bproperty(import_enabled=False)
    longitude = attrs.bproperty(import_enabled=False)
    address = attrs.bproperty(import_enabled=False)
    attending = attrs.bproperty(import_enabled=False)
    in_session = attrs.bproperty(import_enabled=False)
    num_hearing_aid_requested = attrs.bproperty(import_enabled=False)

    # @num_hearing_aid_requested.create_accumulator
    # def num_hearing_aid_request(self, value=_NA):
    #     if value is _NA:
    #         return attrs.aggregation.Accumulator(value=0)
    #     else:
    #         return attrs.aggregation.Accumulator(value=value)

    # @num_hearing_aid_requested.getter
    # def num_hearing_aid_requested(self):
    #     res = 0
    #     for user in self.store.users:
    #         if user.hearing_aid_requested:
    #             res += 1
    #     return res

    @property
    def meeting_id(self):
        return self.store.meeting.doc_id

    @property
    def _view_refs(self):
        for user_id, user in self.store.users.items():
            doc_ref = user.doc_ref / self.__class__.__name__ / self.store.meeting.doc_id
            yield doc_ref

    @in_session.getter
    def in_session(self):
        return self.store.meeting.status == "in-session"

    @in_session.setter
    def in_session(self, in_session):
        cur_status = self.store.meeting.status
        if cur_status == "in-session" and not in_session:
            self.store.meeting.status = "closed"
        elif cur_status == "closed" and in_session:
            self.store.meeting.status = "in-session"
        else:
            raise ValueError

    @latitude.getter
    def latitude(self):
        return self.store.location.latitude

    @longitude.getter
    def longitude(self):
        return self.store.location.longitude

    @address.getter
    def address(self):
        return self.store.location.address

    @attending.getter
    def attending(self):
        user_ids = [uid for uid in self.store.users.keys()]

        if self.store.meeting.status == "not-started":
            return list()

        res = list()
        for user_id in sorted(user_ids):
            ticket = self.store.tickets[user_id]
            user = self.store.users[user_id]
            if ticket.attendance:
                d = {
                    "name": user.display_name,
                    "organization": user.organization,
                    "hearing_aid_requested": user.hearing_aid_requested
                }
                res.append(d)

        return res

    @num_hearing_aid_requested.getter
    def num_hearing_aid_requested(self):
        count = 0
        for d in self.attending:
            if d["hearing_aid_requested"]:
                count += 1
        return count

    # @classmethod
    # def get_many_from_query(cls, query_d=None, once=False):
    #     """ Note that once kwarg apply to the snapshot but not the query.
    #
    #     :param query_d:
    #     :param once: attaches a listener to individual snapshots
    #     :return:
    #     """
    #     return [
    #         cls.get_from_meeting_id(meeting_id=obj.doc_id, once=once)
    #         for obj in Meeting.where(**query_d)]

    @classmethod
    def get(cls, doc_id=None, once=True, meeting=None, **kwargs):

        if meeting is None:
            m: Meeting = Meeting.get(doc_id=doc_id)
        else:
            m = meeting

        struct = dict()

        struct["meeting"] = (Meeting, m.doc_id)

        struct['users'] = dict()
        for user_ref in m.users:
            obj_type = User
            user_id = user_ref.id
            struct["users"][user_id] = (obj_type, user_ref.id)

        struct['tickets'] = dict()
        for user_id, ticket_ref in m.tickets.items():
            obj_type = Ticket
            struct["tickets"][user_id] = (obj_type, ticket_ref.id)

        struct["location"] = (Location, m.location.id)

        store = MeetingSessionStore.from_struct(struct)

        obj = super().get(store=store, once=once,
                          **kwargs)  # TODO: fix super() behavior
        # time.sleep(2)  # TODO: delete after implementing sync

        return obj

    @classmethod
    def get_many_from_query(cls, query_d=None, once=False):
        """ Note that once kwarg apply to the snapshot but not the query.

        :param query_d:
        :param once: attaches a listener to individual snapshots
        :return:
        """
        return [
            cls.get(doc_id=obj.doc_id, once=once)
            for obj in Meeting.where(**query_d)]

    def propagate_change(self):
        self.store.propagate_back()

    # def update_vals(self, *args, user_id, **kwargs):
    #     if user_id in self.store.users:
    #         super().update_vals(*args, **kwargs)
    #     else:
    #         raise UnauthorizedError


class MeetingSessionC(MeetingSessionMixin, view_model.ViewModelR):

    class Meta:
        collection_name = 'meeting_sessions'


class MeetingSession(MeetingSessionMixin, view_model.ViewModel):

    pass


class MeetingSessionMutation(PatchMutation):

    view_model_cls = MeetingSession



from celery import Task
from flasgger import SwaggerView
from flask import request
from google.cloud.firestore import DocumentSnapshot

from onto.context import Context as CTX


def _retrofit(view: SwaggerView, collection_path: str):
    """ Registers a resource so that it is stored in Firestore.

    :return:
    """
    t: Task = CTX.celery_app.task(_pull)
    t.run(view=view, collection_path=collection_path)
    # CTX.celery_app.add_periodic_task(
    #     schedule=1.0,
    #     sig=t
    # )


def _pull(view: SwaggerView, collection_path: str):
    """ Stores REST API response in Firestore.

    :return:
    """
    res = view.get()
    collection = CTX.db.collection(collection_path)
    for key, val in res.items():
        collection.document(document_id=key).set(document_data=val)


def _clear(collection_path: str):
    """ Deletes everything in a collection. USE CAREFULLY!
    NOTE that production release will NOT include this function.

    :param collection_path:
    :return:
    """
    collection = CTX.db.collection(collection_path)
    for doc in collection.stream():
        assert isinstance(doc, DocumentSnapshot)
        doc.reference.delete()




#
#
#
# if __name__ == "__main__":
#
#     coordinator = Coordinator()
#     coordinator.start()
#
#     class MyListener(Listener):
#
#         _coordinator = coordinator
#
#         @classmethod
#         def start(cls):
#             while True:
#                 import time
#                 time.sleep(2)
#                 cls._coordinator._on_snapshot(dict(foo='bar'))
#
#     k = MyListener\
#     k.start()
#
#     raise ValueError



from onto.source.base import Source
from onto.context import Context as CTX


class LeancloudHook(Source):

    def __init__(self, class_name):
        super().__init__()
        self.class_name = class_name

    def start(self):
        self._register()

    def _register(self):
        engine = CTX.services.engine
        for trigger_name in self.protocol.mapping:
            wrapper_f = getattr(engine, trigger_name)
            wrapper = wrapper_f(self.class_name)

            @wrapper
            def f(*args, **kwargs):
                from functools import partial
                f = partial(self._call, trigger_name=trigger_name)
                return f(*args, **kwargs)

    def _call(self, cla_obj, *, trigger_name):
        from onto.database.leancloud import LeancloudSnapshot, LeancloudReference
        snapshot = LeancloudSnapshot.from_cla_obj(cla_obj)
        ref = LeancloudReference.from_cla_obj(cla_obj)
        self._invoke_mediator(
            func_name=trigger_name,
            ref=ref,
            snapshot=snapshot
        )


class DomainModelSource(LeancloudHook):

    def __init__(self, domain_model_cls):
        super().__init__(class_name=domain_model_cls._get_collection_name())
        self.domain_model_cls = domain_model_cls

    def _call(self, trigger_name, cla_obj):
        from onto.database.leancloud import LeancloudSnapshot

        from onto.database.leancloud import LeancloudReference
        ref = LeancloudReference.from_cla_obj(cla_obj)
        snapshot = LeancloudSnapshot.from_cla_obj(cla_obj)
        obj = self.domain_model_cls.from_snapshot(ref=ref, snapshot=snapshot)
        self._invoke_mediator(
            func_name=trigger_name,
            obj=obj
        )

# before_save = LeancloudBeforeSaveSource
hook = LeancloudHook
domain_model_hook = DomainModelSource



from onto.primary_object import PrimaryObject


class DomainModel(PrimaryObject):
    """
    Domain model is intended for handling business logic.
    """

    @classmethod
    def _datastore(cls):
        from onto.context import Context as CTX
        return CTX.db





from onto.models.utils import _schema_cls_from_attributed_class
from onto.registry import ModelRegistry


# DEFAULT_FIELDS = {"obj_type", "doc_id", "doc_ref"}


class AttributedMeta(type):
    @classmethod
    def __collect_attributes(mcs, bases, attrs) -> None:
        attrs['__attributes'] = dict()

        for base in bases:
            if isinstance(base, AttributedMeta):
                base_attributes = base.__dict__['__attributes']
                attrs['__attributes'] = dict(
                    list(attrs['__attributes'].items()) +
                    list(base_attributes.items())  # TODO: debug use case for overridden attribute from super
                )

        def is_attribute(a):
            from onto.attrs.attribute_new import AttributeBase
            return isinstance(a, AttributeBase) and not a.properties.is_root

        for attr_name, a in list(attrs.items()):
            if is_attribute(a):
                attrs['__attributes'][attr_name] = a
                del attrs[attr_name]

        attrs['x'] = lambda attr_name: attrs['__attributes'][attr_name]

    def __new__(mcs, name, bases, attrs):
        mcs.__collect_attributes(bases, attrs)
        klass = super().__new__(mcs, name, bases, attrs)
        return klass


class SerializableMeta(AttributedMeta, ModelRegistry):
    """
    Metaclass for serializable models.
    """

    def __new__(mcs, name, bases, attrs):
        klass = super().__new__(mcs, name, bases, attrs)
        meta = attrs.get('Meta', None)
        if hasattr(meta, "schema_cls"):
            klass._schema_cls = meta.schema_cls
            # TODO: note difference without the else clause
        # else:
        #     klass._schema_cls = None
        if union := getattr(meta, "union", None):
            klass._union = meta.union
        else:
            klass._union = None
        return klass

    def __init__(klass, name, base, ns):
        super().__init__(name, base, ns)

        __attributes = ns['__attributes']

        for attr_name, attr in __attributes.items():
            __attributes[attr_name] = attr.attribute_name(attr_name).parent_klass(parent=klass)

        for attr_name, attr in __attributes.items():
            from onto.attrs.unit import MonadContext
            with MonadContext.context().getter().setter().deleter():
                # print(attr.properties.data_key)
                p = property(
                    fget=getattr(attr, 'fget', None),
                    fset=getattr(attr, 'fset', None),
                    fdel=getattr(attr, 'fdel', None),
                    doc=getattr(attr, 'doc', None) or getattr(attr.properties, 'data_key', None)
                )
                setattr(klass, attr_name, p)


        # if hasattr(klass, "Meta"):
        #     Moves Model.Meta.schema_cls to Model._schema_cls


            # if hasattr(meta, "default_fields"):
            #     default_fields = meta.default_fields
            #     if default_fields > DEFAULT_FIELDS:
            #         raise ModelDeclarationError(
            #             f"default_fields argument: {default_fields} "
            #             f"in {name}.Meta is not a subset (<=) of "
            #             f"{DEFAULT_FIELDS} "
            #         )
            #     klass._default_fields = default_fields





from onto.context import Context as CTX
from onto.config import Config

from onto import schema, fields, utils

from onto.firestore_object import FirestoreObject, \
    ClsFactory
from onto.primary_object import PrimaryObject


if not CTX._ready:
    config = Config(
        app_name="flask-boiler-testing",
        debug=True,
        testing=True,
        certificate_filename="flask-boiler-testing-firebase-adminsdk-4m0ec-7505aaef8d.json"
    )
    CTX.read(config)

    assert CTX.firebase_app.project_id == "flask-boiler-testing"

# Creates a schema for serializing and deserializing to firestore database
class TestObjectSchema(schema.Schema):
    # Describes how obj.int_a is read from and stored to a document in firestore
    int_a = fields.Raw()
    int_b = fields.Raw()


# Declares the object
TestObject = ClsFactory.create(
    name="TestObject",
    schema=TestObjectSchema,
    base=PrimaryObject
)

# Creates an object with default values with reference: "TestObject/testObjId1"
#   (but not saved to database)
obj = TestObject.new(doc_id="testObjId1")
assert obj.doc_id == "testObjId1"
assert obj.collection_name == "TestObject"

# Assigns value to the newly created object
obj.int_a = 1
obj.int_b = 2

# Saves to firestore "TestObject/testObjId1"
obj.save()

# Gets the object from firestore "TestObject/testObjId1"
retrieved_obj = TestObject.get(doc_id="testObjId1")

# Access values of the object retrieved
assert retrieved_obj.int_a == 1
assert retrieved_obj.int_b == 2

# Deletes the object from firestore "TestObject/testObjId1"
retrieved_obj.delete()



from examples.meeting_room.domain_models.user import User
from examples.meeting_room.domain_models.meeting import Meeting
from examples.meeting_room.view_models import MeetingSession
from onto import view_model
from onto.attrs import attrs
from onto.context import Context as CTX
from onto.store import Store, reference


class UserStore(Store):
    user = reference(dm_cls=User)


class UserViewMixin:

    user_id: str = attrs.of_type(str).optional
    first_name = attrs.str.data_key('firstName')
    last_name = attrs.str.data_key('lastName')
    organization = attrs.str.data_key('organization')

    hearing_aid_requested = attrs.bool.data_key('hearingAidRequested')
    # meetings = attrs.bproperty(import_enabled=False)

    @first_name.getter
    def first_name(self):
        return self.store.user.first_name

    @last_name.getter
    def last_name(self):
        return self.store.user.last_name

    @last_name.setter
    def last_name(self, new_last_name):
        self.store.user.last_name = new_last_name

    @organization.getter
    def organization(self):
        return self.store.user.organization

    @hearing_aid_requested.getter
    def hearing_aid_requested(self):
        return self.store.user.hearing_aid_requested
    #
    # @meetings.getter
    # def meetings(self):
    #     meetings_generator = Meeting.where(
    #         users=("array_contains", str(self.store.user.doc_ref))
    #     )
    #     return [
    #         MeetingSession.get(meeting=meeting).to_dict()
    #         for meeting in meetings_generator
    #     ]

    @classmethod
    def get(cls, user_id, once=False, **kwargs):
        struct = dict()
        struct["user"] = (User, user_id)
        store = UserStore.from_struct(struct)
        return super().get(store=store, once=once, user_id=user_id,
                           **kwargs)

    get_from_user_id = get


class UserView(UserViewMixin, view_model.ViewModel):
    pass


class UserViewDAV(UserViewMixin, view_model.ViewModel):

    @property
    def doc_ref(self):
        doc_ref = CTX.db.ref/"UserViewDAV"/self.user_id
        return doc_ref



"""

reuse graphene-pydantic code to convert flask-boiler field to
graphene field.

Ref: https://github.com/graphql-python/graphene-pydantic/blob/master/graphene_pydantic/converters.py

   Copyright 2019 The Upside Travel Company LLC. All Rights Reserved.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import collections
import collections.abc
import datetime
import decimal
import enum
import sys
import typing as T
import uuid

from graphql.type.scalars import (
    GraphQLID,
    GraphQLBoolean,
    GraphQLString,
    GraphQLFloat,
    GraphQLInt,
    GraphQLNamedType
)
# from pydantic import BaseModel
# from pydantic.fields import ModelField
#
# from .registry import Registry
# from .util import construct_union_class_name


# Pydantic 1.0+
# from pydantic import fields

# SHAPE_SINGLETON = (fields.SHAPE_SINGLETON,)
# SHAPE_SEQUENTIAL = (
#     fields.SHAPE_LIST,
#     fields.SHAPE_TUPLE,
#     fields.SHAPE_TUPLE_ELLIPSIS,
#     fields.SHAPE_SEQUENCE,
#     fields.SHAPE_SET,
# )
# SHAPE_MAPPING = (fields.SHAPE_MAPPING,)


NONE_TYPE = None.__class__  # need to do this because mypy complains about type(None)
#

class ConversionError(TypeError):
    pass


def get_attr_resolver(attr_name: str) -> T.Callable:
    """
    Return a helper function that resolves a field with the given name by
    looking it up as an attribute of the type we're trying to resolve it on.
    """

    def _get_field(root, _info):
        return getattr(root, attr_name, None)

    return _get_field


from onto.attrs.attribute import AttributeBase


def convert_fb_input_field(
    field: AttributeBase,
    **field_kwargs,
) -> InputField:
    """
    Convert a Pydantic model field into a Graphene type field that we can add
    to the generated Graphene data model type.
    """
    declared_type = getattr(field, "type_", None)
    field_kwargs.setdefault(
        "type",
        convert_pydantic_type(
            declared_type, field, registry, parent_type=parent_type, model=model
        ),
    )
    field_kwargs.setdefault("required", field.required)
    field_kwargs.setdefault("default_value", field.default)
    # TODO: find a better way to get a field's description. Some ideas include:
    # - hunt down the description from the field's schema, or the schema
    #   from the field's base model
    # - maybe even (Sphinx-style) parse attribute documentation
    field_kwargs.setdefault("description", field.__doc__)

    return InputField(**field_kwargs)


def convert_pydantic_field(
    field: ModelField,
    registry: Registry,
    parent_type: T.Type = None,
    model: T.Type[BaseModel] = None,
    **field_kwargs,
) -> Field:
    """
    Convert a Pydantic model field into a Graphene type field that we can add
    to the generated Graphene data model type.
    """
    declared_type = getattr(field, "type_", None)
    field_kwargs.setdefault(
        "type",
        convert_pydantic_type(
            declared_type, field, registry, parent_type=parent_type, model=model
        ),
    )
    field_kwargs.setdefault("required", field.required)
    field_kwargs.setdefault("default_value", field.default)
    # TODO: find a better way to get a field's description. Some ideas include:
    # - hunt down the description from the field's schema, or the schema
    #   from the field's base model
    # - maybe even (Sphinx-style) parse attribute documentation
    field_kwargs.setdefault("description", field.__doc__)

    return Field(resolver=get_attr_resolver(field.name), **field_kwargs)


def convert_pydantic_type(
    attr_cls,
) -> BaseType:  # noqa: C901
    """
    Convert a Pydantic type to a Graphene Field type, including not just the
    native Python type but any additional metadata (e.g. shape) that Pydantic
    knows about.
    """

    TYPE_MAP = {

    }

    graphene_type = find_graphene_type(
        type_, field, registry, parent_type=parent_type, model=model
    )
    if field.shape in SHAPE_SINGLETON:
        return graphene_type
    elif field.shape in SHAPE_SEQUENTIAL:
        # TODO: _should_ Sets remain here?
        return List(graphene_type)
    elif field.shape in SHAPE_MAPPING:
        raise ConversionError("Don't know how to handle mappings in Graphene.")


def find_graphene_type(
    type_: T.Type,
    field: ModelField,
    registry: Registry,
    parent_type: T.Type = None,
    model: T.Type[BaseModel] = None,
) -> BaseType:  # noqa: C901
    """
    Map a native Python type to a Graphene-supported Field type, where possible,
    throwing an error if we don't know what to map it to.
    """
    if type_ == uuid.UUID:
        return UUID
    elif type_ in (str, bytes):
        return String
    elif type_ == datetime.datetime:
        return DateTime
    elif type_ == datetime.date:
        return Date
    elif type_ == datetime.time:
        return Time
    elif type_ == bool:
        return Boolean
    elif type_ == float:
        return Float
    elif type_ == decimal.Decimal:
        return GrapheneDecimal if DECIMAL_SUPPORTED else Float
    elif type_ == int:
        return Int
    elif type_ in (tuple, list, set):
        # TODO: do Sets really belong here?
        return List
    elif registry and registry.get_type_for_model(type_):
        return registry.get_type_for_model(type_)
    elif registry and isinstance(type_, BaseModel):
        # If it's a Pydantic model that hasn't yet been wrapped with a ObjectType,
        # we can put a placeholder in and request that `resolve_placeholders()`
        # be called to update it.
        registry.add_placeholder_for_model(type_)
    # NOTE: this has to come before any `issubclass()` checks, because annotated
    # generic types aren't valid arguments to `issubclass`
    elif hasattr(type_, "__origin__"):
        return convert_generic_python_type(
            type_, field, registry, parent_type=parent_type, model=model
        )
    elif isinstance(type_, T.ForwardRef):
        # A special case! We have to do a little hackery to try and resolve
        # the type that this points to, by trying to reference a "sibling" type
        # to where this was defined so we can get access to that namespace...
        sibling = model or parent_type
        if not sibling:
            raise ConversionError(
                "Don't know how to convert the Pydantic field "
                f"{field!r} ({field.type_}), could not resolve "
                "the forward reference. Did you call `resolve_placeholders()`? "
                "See the README for more on forward references."
            )
        module_ns = sys.modules[sibling.__module__].__dict__
        resolved = type_._evaluate(module_ns, None)
        # TODO: make this behavior optional. maybe this is a place for the TypeOptions to play a role?
        if registry:
            registry.add_placeholder_for_model(resolved)
        return find_graphene_type(
            resolved, field, registry, parent_type=parent_type, model=model
        )
    elif issubclass(type_, enum.Enum):
        return Enum.from_enum(type_)
    elif issubclass(type_, str):
        return String
    else:
        raise ConversionError(
            f"Don't know how to convert the Pydantic field {field!r} ({field.type_})"
        )


def convert_generic_python_type(
    type_: T.Type,
    field: ModelField,
    registry: Registry,
    parent_type: T.Type = None,
    model: T.Type[BaseModel] = None,
) -> BaseType:  # noqa: C901
    """
    Convert annotated Python generic types into the most appropriate Graphene
    Field type -- e.g. turn `typing.Union` into a Graphene Union.
    """
    origin = type_.__origin__
    if not origin:  # pragma: no cover  # this really should be impossible
        raise ConversionError(f"Don't know how to convert type {type_!r} ({field})")

    # NOTE: This is a little clumsy, but working with generic types is; it's hard to
    # decide whether the origin type is a subtype of, say, T.Iterable since typical
    # Python functions like `isinstance()` don't work
    if origin == T.Union:
        return convert_union_type(
            type_, field, registry, parent_type=parent_type, model=model
        )
    elif origin in (
        T.Tuple,
        T.List,
        T.Set,
        T.Collection,
        T.Iterable,
        list,
        set,
    ) or issubclass(origin, collections.abc.Sequence):
        # TODO: find a better way of divining that the origin is sequence-like
        inner_types = getattr(type_, "__args__", [])
        if not inner_types:  # pragma: no cover  # this really should be impossible
            raise ConversionError(
                f"Don't know how to handle {type_} (generic: {origin})"
            )
        # Of course, we can only return a homogeneous type here, so we pick the
        # first of the wrapped types
        inner_type = inner_types[0]
        return List(
            find_graphene_type(
                inner_type, field, registry, parent_type=parent_type, model=model
            )
        )
    elif origin in (T.Dict, T.Mapping, collections.OrderedDict, dict) or issubclass(
        origin, collections.abc.Mapping
    ):
        raise ConversionError("Don't know how to handle mappings in Graphene")
    else:
        raise ConversionError(f"Don't know how to handle {type_} (generic: {origin})")


def convert_union_type(
    type_: T.Type,
    field: ModelField,
    registry: Registry,
    parent_type: T.Type = None,
    model: T.Type[BaseModel] = None,
):
    """
    Convert an annotated Python Union type into a Graphene Union.
    """
    inner_types = type_.__args__
    # We use a little metaprogramming -- create our own unique
    # subclass of graphene.Union that knows its constituent Graphene types
    parent_types = tuple(
        find_graphene_type(x, field, registry, parent_type=parent_type, model=model)
        for x in inner_types
        if x != NONE_TYPE
    )

    # This is effectively a typing.Optional[T], which decomposes into a
    # typing.Union[None, T] -- we can return the Graphene type for T directly
    # since Pydantic will have already parsed it as optional
    if len(parent_types) == 1:
        return parent_types[0]

    internal_meta_cls = type("Meta", (), {"types": parent_types})

    union_cls = type(
        construct_union_class_name(inner_types), (Union,), {"Meta": internal_meta_cls}
    )
    return union_cls


# from google.cloud.firestore import DocumentReference
import abc

# from google.cloud.firestore_v1 import WriteOption, LastUpdateOption

from onto.common import _NA
from onto.database import Snapshot, Reference
from onto.mapper.helpers import RelationshipReference
# from onto.view_model import ViewModel
from onto.models.mixin import resolve_obj_cls
from onto.registry import ModelRegistry
from onto.models.base import Serializable
from onto.utils import snapshot_to_obj
from onto.context import Context as CTX


class FirestoreObjectMixin:

    def __init__(self, *args, doc_ref=None, transaction=_NA, **kwargs):

        if transaction is _NA:
            transaction = CTX.transaction_var.get()

        self._doc_ref = doc_ref
        self.transaction = transaction

        super().__init__(*args, **kwargs)

    # def get_firestore_ref(self):
    #     warnings.warn("Please use .doc_ref instead. ", DeprecationWarning)
    #     return self.doc_ref

    @classmethod
    @abc.abstractmethod
    def _datastore(cls):
        pass

    @property
    def doc_ref(self) -> Reference:
        """ Returns the Document Reference of this object.
        """
        return self._doc_ref

    @classmethod
    def get(cls, *, doc_ref=None, transaction=_NA, **kwargs):
        """ Retrieves an object from Firestore

        :param doc_ref:
        :param transaction:
        :param kwargs: Keyword arguments to be forwarded to from_dict
        """

        snapshot = cls._datastore().get(ref=doc_ref, transaction=transaction)
        obj = snapshot_to_obj(
            snapshot=snapshot,
            reference=doc_ref,
            super_cls=cls,
            transaction=transaction)
        return obj

    @classmethod
    def from_snapshot(cls, ref, snapshot=None, **kwargs):
        """ Deserializes an object from a Document Snapshot.

        :param snapshot: Firestore Snapshot
        """
        # if not snapshot.exists:
        #     return None

        obj = cls.from_dict(d=snapshot.to_dict(), doc_ref=ref, **kwargs)
        return obj

    def to_snapshot(self):
        return Snapshot(**self.to_dict())

    def save(self,
             transaction: 'google.cloud.firestore.Transaction'=_NA,
             doc_ref=None,
             _store=_NA,
             ):
        """ Save an object to Firestore

        :param transaction: Firestore Transaction
        :param doc_ref: override save with this doc_ref
        :param save_rel: If true, objects nested in this
            object will be saved to the Firestore.
        """

        if transaction is _NA:
            transaction = CTX.transaction_var.get()

        if doc_ref is None:
            doc_ref = self.doc_ref

        if _store is _NA:
            if issubclass(self.__class__, FirestoreObjectValMixin):
                _store = self._store
            else:
                _store = None

        d = self._export_as_dict(transaction=transaction, _store=_store)
        _store.save()
        snapshot = Snapshot(d)
        self._datastore().set(snapshot=snapshot, ref=doc_ref, transaction=transaction)

    def delete(self, transaction: 'google.cloud.firestore.Transaction' = _NA) -> None:
        """ Deletes and object from Firestore.

        :param transaction: Firestore Transaction
        """

        if transaction is _NA:
            transaction = CTX.transaction_var.get()

        self._datastore().delete(ref=self.doc_ref, transaction=transaction)


def _nest_relationship_import(rr, _store):
    """ (Experimental)

    :param rr:
    :param container:
    :return:
    """
    if rr.obj is not None:
        return rr.obj

    doc_ref, obj_type = rr.doc_ref, rr.obj_type
    if isinstance(obj_type, str):
        obj_type = ModelRegistry.get_cls_from_name(obj_type)

    _store.insert(doc_ref=doc_ref, obj_type=obj_type)
    from peak.util.proxies import CallbackProxy
    import functools
    _callback = functools.partial(_store.retrieve, doc_ref=doc_ref, obj_type=obj_type)
    return CallbackProxy(_callback)


# def _nest_relationship_export(rr, store):
#     """ (Experimental)
#
#     :param rr:
#     :param container:
#     :return:
#     """
#     obj, obj_type = rr.obj, rr.obj_type
#
#     store.insert(doc_ref=doc_ref, obj_type=obj_type)
#     from peak.util.proxies import CallbackProxy
#     import functools
#     _callback = functools.partial(store.retrieve, doc_ref=doc_ref, obj_type=obj_type)
#     return CallbackProxy(_callback)


# def _get_snapshots(transaction, **kwargs):
#     """ needed because transactional wrapper uses specific argument
#         ordering
#
#     :param transaction:
#     :param kwargs:
#     :return:
#     """
#     return CTX.db.get_many(transaction=transaction, **kwargs)


class FirestoreObjectValMixin:
    """
    Subclasses of this class can hold FirestoreObject as an attribute
    """

    @classmethod
    @abc.abstractmethod
    def _datastore(cls):
        """
        NOTE: this is declared in all of:
            - FirestoreObjectMixin
            - FirestoreObjectValMixin
        :return:
        """
        pass

    def __init__(self, *args, _store=_NA, **kwargs):

        if _store is _NA:
            """ NOTE: This branch should only be invoked for objects that are 
                created with new and never retrieved from datastore.
            """
            from onto.store import Gallery
            _store = Gallery()

        self._store = _store
        super().__init__(*args, **kwargs)

    def _export_val(self, val, transaction=None, _store=None):

        def nest_relationship(obj):
            if _store is None:
                # TODO: change
                obj.save(transaction=transaction)
            else:
                _store.save_later(obj, transaction=transaction)

            return str(obj.doc_ref)

        if isinstance(val, RelationshipReference):
            if val.nested and val.obj is not None:
                return nest_relationship(val.obj)
            elif val.nested and val.doc_ref is not None:
                return val.doc_ref
            elif not val.nested and val.doc_ref is not None:
                return val.doc_ref
            elif not val.nested and val.obj is not None:
                """ We should stay away from serializing a 
                    referenced object in domain model since references on 
                    domain model can be circularly referenced. Such case 
                    results in a loop that may create dangerous amount of 
                    external or internal API usage. 
                """
                raise ValueError
            else:
                return None
        else:
            return super()._export_val(val, transaction=transaction, _store=_store)


    # def _export_val_view(self, val):
    #     def is_nested_relationship(val):
    #         return isinstance(val, RelationshipReference) and val.nested
    #
    #     def is_ref_only_relationship(val):
    #         return isinstance(val, RelationshipReference) and not val.nested
    #
    #     if is_nested_relationship(val):
    #         return super()._export_val_view(val.obj)
    #     elif is_ref_only_relationship(val):
    #         return super()._export_val_view(val.doc_ref.path)
    #     else:
    #         return super()._export_val_view(val)

    def _export_as_dict(self, transaction=None, _store=_NA, **kwargs):

        d = self.schema_obj.dump(self)

        if _store is _NA:
            from onto.store import Gallery
            _store = self._store
            res = self._export_val(d, transaction=transaction, _store=_store,
                                   **kwargs)
            _store.refresh(transaction=transaction)
        else:
            res = self._export_val(d, transaction=transaction, _store=_store,
                                   **kwargs)

        return res

    # def _export_val_view(self, val):
    #
    #     def get_vm(doc_ref):
    #         obj = FirestoreObject.get(doc_ref=doc_ref,
    #                                   transaction=self.transaction)
    #         return obj._export_as_view_dict()
    #
    #     if isinstance(val, RelationshipReference):
    #         if val.obj is not None:
    #             return val.obj._export_as_view_dict()
    #         elif val.doc_ref is not None:
    #             return get_vm(val.doc_ref)
    #         else:
    #             return val.doc_ref
    #     else:
    #         return super()._export_val_view(val)

    @classmethod
    def _import_val(cls, val, transaction=None, partial=False, _store=None):
        # TODO: note partial is not applied to _nest_relationship_import calls
        if isinstance(val, RelationshipReference):
            if val.nested and val.doc_ref is not None:
                return _nest_relationship_import(val, _store=_store)
            elif not val.nested and val.doc_ref is not None:
                return val.doc_ref
            elif val.nested and val.obj is not None:
                return _nest_relationship_import(val, _store=_store)
            else:
                raise ValueError
        else:
            return super()._import_val(val, transaction=transaction, partial=partial, _store=_store)

    @classmethod
    def _import_from_dict(cls, d, transaction=None, _store=_NA, **kwargs):
        if _store is _NA:
            from onto.store import Gallery
            _store = Gallery()
            res = cls._import_val(d, transaction=transaction, _store=_store, **kwargs)
            _store.refresh(transaction=transaction)
        else:
            res = cls._import_val(d, transaction=transaction, _store=_store,
                                  **kwargs)
        res['_store'] = _store
        return res

    @classmethod
    def from_dict(
            cls,
            d,
            _store=_NA,
            transaction=None,
            partial=False,
            **kwargs):
        """ Deserializes an object from a dictionary.

        :param d: a dictionary representation of an object generated
            by `to_dict` method.
        :param transaction: Firestore transaction for retrieving
            related documents, and for saving this object.
        :param kwargs: Keyword arguments to be forwarded to new
        """

        obj_cls = resolve_obj_cls(cls=cls, d=d)

        schema_obj = obj_cls.get_schema_obj()
        d = schema_obj.load(d, partial=partial)

        d = cls._import_from_dict(d, _store=_store, partial=partial, transaction=transaction)

        instance = obj_cls.new(**d, **kwargs)
        # TODO: fix unexpected arguments
        return instance


class FirestoreObject(
    FirestoreObjectValMixin,
                      FirestoreObjectMixin,
                      Serializable):

    @classmethod
    def _datastore(cls):
        _db = CTX.db
        return _db



from pony.orm import db_session, sql_debug, select, count
from pony.orm.core import Entity, left_join, make_query
from examples.meeting_room.domain_models.location import Location

from onto import attrs
from onto.collection_mixin import db
from pony.orm.dbproviders.sqlite import SQLiteProvider


def f(p):
    return 'P'


def g(q):
    return 'Q'


from pony.orm import raw_sql


class InsertMixin:

    @classmethod
    def select(cls):
        from onto.models.utils import _collect_attrs
        from onto.attrs.attribute import PropertyAttributeBase

        def yk(item):
            for key, attribute in _collect_attrs(cls):
                # if not attribute.export_enabled:
                #     continue
                    # else:
                    #     res = 'one'
                    # return res
                        # return attribute.__get__(item, item.__class__).fget()
                    # else:
                    #     return 'one'
                        # return getattr(item, key)
                if isinstance(attribute, PropertyAttributeBase) and attribute.fget is not None:
                    def f(item):
                        return getattr(cls, key).fget(item)
                else:
                    def f(item):
                        return getattr(item, key)
                yield f(item)

        def f(item, key, cls):
            return getattr(cls, key).fget(item)

        def g(item, key):
            return getattr(item, key)

        def yk(item, cls):
            return (f(item, 'ct', cls), f(item, 'latlng', cls), f(item, 'status', cls))

        # make_query()
        gen = select(item for item in cls)
        # cls.ct.fget(item),
        res = select(yk(item, cls) for item in gen).get_sql()
        print(res)

    @classmethod
    def _insert_dynamic_table(entity):


        def f(x, y):
            return x, y
            # avdict = {}
            # for attr in entity._attrs_:
            #     # import functools
            #     monad = getattr(entity, attr.name).(x, y)
            #     # p = functools.partial(monad, owner=entity)
            #     # res = select(getattr(m, attr.name) for m in Meeting)
            #     avdict[attr] = monad(x)
            # if entity._pk_is_composite_:
            #     from pony.py23compat import imap
            #     pkval = tuple(imap(avdict.get, entity._pk_attrs_))
            #     if None in pkval: pkval = None
            # else:
            #     pkval = avdict.get(entity._pk_attrs_[0])
            #
            # d = dict()
            # for attr in avdict:
            #     # TODO: support columns
            #     val = avdict[attr]
            #     if val is None: continue
            #     d[attr.name] = val
            #
            # table_name = entity._table_  # TODO: may need to do db.get_table(entity._table_)


        def op(x, y):
            return (x.status, y)

        def g():
            return (
                (x,y)
                for x in MeetingTwo
                for y in x.tickets
            )

        def query_gen(f, g):
            return select( f(x, y) for x,y in g() )

        def fg(m):
            return m.location.latitude

        def lng(m):
            return m.location.longitude

        im = query_gen(f=op, g=g)

        from pony.orm.sqltranslation import SqlQuery
        ast = im._translator.construct_subquery_ast(aliases=['store.status', 'store.ticket'])
        sql, adapter = db._ast2sql(ast)

        # from pony.py23compat import values_list
        #
        # arguments = adapter(values_list(kw))
        # return sql
        return sql


from examples.meeting_room.domain_models import Meeting

class MeetingTwo(Meeting, InsertMixin):

    ct = attrs.integer(type_cls=int)
    latlng = attrs.string(type_cls=str)

    @latlng.getter
    def latlng(self):
        return str(self.location.latitude) + str(self.location.longitude)

    @Meeting.status.getter
    def status(self):
        return 's'

    @ct.getter
    def ct(self):
        return sum(count(1) for t in self.tickets if t.attendance)



# class MeetingSession:
#
#     @row_iter
#     def meetings(cls):
#         """ yields the source of each row
#
#         :return:
#         """
#         for m in Meeting:
#             for t in Ticket:
#                 yield Store(
#                     meeting=m,
#                     ticket=t,
#                 )


@db_session
def meeting():


    # m = Meeting.new(doc_id='ab', )
    # m.status = "in-session"

    res = MeetingTwo._insert_dynamic_table()
    print(res)
    assert None is not None



    meetings = lambda: (m for m in Meeting)

    def in_session_meetings(meetings):
        for m in meetings:
            if m.status == 'in-session':
                yield m

    def cti(meeting):
        return meeting.status == 'in-session'

    # def ct(meetings):
    #     for m in meetings:
    #         yield cti()

    s = select(cti(m) for m in meetings()).get_sql()
    print(s)

if __name__ == "__main__":

    import pony.options
    pony.options.CUT_TRACEBACK = False

    from examples.meeting_room.domain_models import Meeting, Location, User, Ticket
    sql_debug(True)  # Output all SQL queries to stdout

    db.bind(SQLiteProvider, filename=':memory:', create_db=True)
    db.generate_mapping(check_tables=False, create_tables=True)

    m = MeetingTwo.new(doc_id='m2')
    m.save()

    # MeetingTwo.select()
    #
    # res = select( (m.latlng, m.status) for m in MeetingTwo)
    # res = res.get_sql()
    # print(res)
    # meeting()
    #
    # assert False



from statefun import make_json_type, Context, Message, kafka_egress_message
from onto.models.base import Serializable

METHOD_TYPE = make_json_type(typename="example/Method")
EVENT_TYPE = make_json_type(typename="example/Event")

EGRESS_RECORD_TYPE = make_json_type(typename="io.statefun.playground/EgressRecord")


async def do_classmethod(obj_cls: type, classmethod_call):
    function_name = classmethod_call['f']
    parameters = classmethod_call['parameters']
    for k, v in classmethod_call.get('serializable_parameters', dict()).items():
        parameters[k] = Serializable.from_dict(v)
    f = getattr(obj_cls, function_name)
    return f(**parameters)


async def do_init(obj_cls: type, method_call: dict, storage: Context.storage) -> None:
    res = await do_classmethod(obj_cls, classmethod_call=method_call)
    d: dict = res.to_dict()
    import json
    __d = json.dumps(d)
    storage.__d = __d


async def do_event(obj_cls: type, event: dict, storage: Context.storage, doc_id: str) -> None:
    __d = storage.__d
    if not __d:
        raise TypeError("NULL 未初始化的对象不可调用 do_event")
    import json
    d: dict = json.loads(__d)
    obj = obj_cls.from_dict(d)

    name: str = event['name']  # name = 'rule-hello-world'
    function_name = '_'.join(name.split('-')[1:])  # function_name = 'hello_world'
    f = getattr(obj, function_name)
    import inspect
    if inspect.iscoroutinefunction(f):
        await f(event)
    else:
        f(event)

    d: dict = obj.to_dict()
    import json
    __d = json.dumps(d)
    storage.__d = __d


async def do_method(obj_cls: type, method_call: dict, storage: Context.storage, doc_id: str) -> None:
    __d = storage.__d
    if not __d:
        raise TypeError("NULL 未初始化的对象不可调用 method")
    import json
    d: dict = json.loads(__d)
    obj = obj_cls.from_dict(d)

    function_name = method_call['f']
    parameters = method_call['parameters']
    for k, v in method_call.get('serializable_parameters', dict()).items():
        parameters[k] = Serializable.from_dict(v)
    f = getattr(obj, function_name)
    import inspect
    if inspect.iscoroutinefunction(f):
        await f(**parameters)
    else:
        f(**parameters)

    d: dict = obj.to_dict()
    import json
    __d = json.dumps(d)
    storage.__d = __d


async def do_raw(obj_cls: type, method_call: dict, storage: Context.storage, context: Context, message: Message) -> None:
    """
    context and message passthrough as __context__ and __message__
    """
    __d = storage.__d
    if not __d:
        raise TypeError("NULL 未初始化的对象不可调用 method")
    import json
    d: dict = json.loads(__d)
    obj = obj_cls.from_dict(d)

    function_name = method_call['f']
    parameters = method_call['parameters']
    for k, v in method_call.get('serializable_parameters', dict()).items():
        parameters[k] = Serializable.from_dict(v)
    f = getattr(obj, function_name)
    import inspect
    if inspect.iscoroutinefunction(f):
        await f(**parameters, __context__=context, __message__=message)
    else:
        f(**parameters, __context__=context, __message__=message)

    d: dict = obj.to_dict()
    import json
    __d = json.dumps(d)
    storage.__d = __d


async def make_event_call(dm_cls: type, context: Context, message: Message, topic: str):
    """
    React to event
    """
    if not message.is_type(EVENT_TYPE):
        raise TypeError('需要是 EVENT_TYPE')
    event = message.as_type(EVENT_TYPE)
    await do_event(dm_cls, event=event, storage=context.storage, doc_id=message.target_id)

    context.send_egress(kafka_egress_message(
        typename='com.example/domain-egress',
        topic=topic,
        key=context.address.id,
        value=context.storage.__d)
    )


async def make_call(dm_cls: type, context: Context, message: Message, topic: str):
    if not message.is_type(METHOD_TYPE):
        raise TypeError('需要是 METHOD_TYPE')
    function_call = message.as_type(METHOD_TYPE)
    if function_call['invocation_type'] == 'ClassMethod':
        await do_init(dm_cls, method_call=function_call, storage=context.storage)
    elif function_call['invocation_type'] == 'Method':
        await do_method(dm_cls, method_call=function_call, storage=context.storage, doc_id=message.target_id)
    elif function_call['invocation_type'] == 'DoNoOpEmit':
        """
        emit a view
        """
        pass
    elif function_call['invocation_type'] == 'RawMethod':
        await do_raw(
            dm_cls,
            method_call=function_call,
            storage=context.storage,
            context=context,
            message=message
        )
    else:
        raise
    context.send_egress(kafka_egress_message(
        typename='com.example/domain-egress',
        topic=topic,
        key=context.address.id,
        value=context.storage.__d)
    )


async def send_one(s, topic, target_id: str):
    """ TODO: optimize """
    from aiokafka import AIOKafkaProducer
    producer = AIOKafkaProducer(bootstrap_servers='kafka.kafka.svc.cluster.local:9092')
    # Get cluster layout and initial topic/partition leadership information
    await producer.start()
    try:
        # Produce message
        await producer.send_and_wait(topic, value=s.encode('utf-8'), key=target_id.encode('utf-8'))
    except Exception as e:
        raise
    finally:
        # Wait for all pending messages to be delivered or expire.
        await producer.stop()


class StatefunProxy:

    def __init__(self, wrapped, target_id: str, invocation_type: str, topic: str):
        self.wrapped = wrapped
        self.target_id = target_id
        self.invocation_type = invocation_type
        self.topic = topic

    @classmethod
    def method_of(cls, dm_cls: type, target_id: str):
        return cls(
            wrapped=dm_cls,
            target_id=target_id,
            invocation_type='Method',
            topic=f'{dm_cls.__name__.casefold()}-calls-1'
        )

    @classmethod
    def raw_method_of(cls, dm_cls: type, target_id: str):
        return cls(
            wrapped=dm_cls,
            target_id=target_id,
            invocation_type='RawMethod',
            topic=f'{dm_cls.__name__.casefold()}-calls-1'
        )

    @classmethod
    def classmethod_of(cls, dm_cls: type, target_id: str):
        return cls(
            wrapped=dm_cls,
            target_id=target_id,
            invocation_type='ClassMethod',
            topic=f'{dm_cls.__name__.casefold()}-calls-1'
        )

    @classmethod
    def noopemit_of(cls, dm_cls: type, target_id: str):
        return cls(
            wrapped=dm_cls,
            target_id=target_id,
            invocation_type='NoOpEmit',
            topic=f'{dm_cls.__name__.casefold()}-calls-1'
        )



    def __getattr__(self, name):

        async def make_call(**kwargs):
            print(self.invocation_type)
            print(f'id: {self.target_id} f: {name} parameters: {kwargs}')
            serializable_parameters = {k:v.to_dict() for k, v in kwargs.items() if isinstance(v, Serializable) }
            regular_parameters = {k:v for k, v in kwargs.items() if not isinstance(v, Serializable) }
            res = dict(f=name, parameters=regular_parameters, serializable_parameters=serializable_parameters, invocation_type=self.invocation_type)
            import json
            s = json.dumps(res)
            await send_one(s, self.topic, target_id=self.target_id)

        return make_call





import logging

from onto.source.base import Source


def default_key_deserializer(v: bytes):
    return v.decode("utf-8")


def default_value_deserializer(v: bytes):
    import json
    if v is None:
        """
        tombstone
        """
        return None
    else:
        s = v.decode('utf-8')
        return json.loads(s)


async def _kafka_subscribe(
        *,
        topic_name,
        callback,
        bootstrap_servers='kafka.default.svc.cluster.local:9092',
        silent_errors=(),
        silent_errors_handler=None,
        **kwargs
):
    from aiokafka import AIOKafkaConsumer

    consumer = AIOKafkaConsumer(
        topic_name,
        bootstrap_servers=bootstrap_servers,
        **kwargs
    )
    # Get cluster layout and join group `my-group`
    await consumer.start()

    async def generator():
        """
        To allow errors to occur and handled without terminating the async loop
        """

        async def inner():
            async for msg in consumer:
                yield msg

        while True:
            try:
                async for _msg in inner():
                    yield _msg
                else:
                    break
            except silent_errors as e:
                silent_errors_handler(e)

    # TODO: make first yield when listener has started
    try:
        # Consume messages
        async for msg in generator():
            await callback(message=msg)
            # print("consumed: ", msg.topic, msg.partition, msg.offset,
            #       msg.key, msg.value, msg.timestamp)
    except Exception as e:
        import warnings
        logging.exception('invoke kafka failed')
        raise ValueError('Interrupted') from e
    finally:
        # Will leave consumer group; perform autocommit if enabled.
        await consumer.stop()


class KafkaSource(Source):

    def __init__(
            self,
            topic_name,
            bootstrap_servers='kafka.default.svc.cluster.local:9092',
            key_deserializer=default_key_deserializer,
            value_deserializer=default_value_deserializer,
            group_id=None,
            **kwargs
    ):
        """ Initializes a ViewMediator to declare protocols that
                are called when the results of a query change. Note that
                mediator.start must be called later.

        :param query: a listener will be attached to this query
        """
        super().__init__()  # NOTE: no kwargs call
        self.topic_name = topic_name
        self.bootstrap_servers = bootstrap_servers
        self.kwargs = kwargs
        self.key_deserializer = key_deserializer
        self.value_deserializer = value_deserializer

    def start(self, loop):
        import asyncio
        # tell asyncio to enqueue the result
        _ = asyncio.run_coroutine_threadsafe(
            self._register(), loop=loop
        )

    async def _register(self):
        try:
            from functools import partial
            f = partial(self._invoke_mediator_async, func_name='on_topic')
            await _kafka_subscribe(
                topic_name=self.topic_name,
                callback=f,
                bootstrap_servers=self.bootstrap_servers,
                key_deserializer=self.key_deserializer,
                value_deserializer=self.value_deserializer,
                **self.kwargs
            )
        except Exception as _:
            import logging
            logging.exception(f'async _register failed')



class KafkaDomainModelSource(KafkaSource):

    def __init__(self, *, dm_cls, **kwargs, ):
        self.dm_cls = dm_cls
        super().__init__(**kwargs)

    async def _invoke_mediator_async(self, *, func_name, message: 'ConsumerRecord'):
        k = message.key
        v = message.value
        obj = self.dm_cls.from_dict(v)
        # obj.doc_id = k
        try:
            await super()._invoke_mediator_async(func_name=func_name, obj=obj)
        except Exception as _:
            import logging
            logging.exception(f'async _invoke_mediator failed for {func_name} {str(k)}')





from onto.context import Context as CTX

from google.cloud.firestore import Watch, DocumentSnapshot, \
    DocumentReference, Query

from onto.view.base import ViewMediatorBase


class ViewMediatorDAV(ViewMediatorBase):
    """
    Watches and updates Firestore for DocumentAsView view models
    """

    def __init__(self,
                 view_model_cls=None,
                 mutation_cls=None):
        """ Initializes a Document View Mediator.

        :param view_model_cls: the view model to be exposed with Document.
        :param mutation_cls: a subclass of Mutation to handle the changes
                made to the list of view models or a single view model.
        """
        super().__init__()
        self.view_model_cls = view_model_cls
        self.mutation_cls = mutation_cls
        self.rule_view_cls_mapping = dict()
        if self.view_model_cls is not None:
            self.default_tag = self.view_model_cls.__name__
        else:
            # Add default_tag as HIDDEN for internal hooks
            self.default_tag = "HIDDEN"
        self.instances = dict()

    @classmethod
    def notify(cls, obj):
        """ Specifies what to do with the view model newly generated from
                an update.

        :param obj: view model newly generated
        """
        obj.save()

    def _get_collection_name(self):
        return self.view_model_cls.__name__

    def _get_patch_query(self) -> Query:
        collection_group_id = "_PATCH_{}" \
            .format(self._get_collection_name())
        collection_group_query = CTX.db.collection_group(collection_group_id)
        return collection_group_query

    def _listen_to_patch(self):
        # NOTE: index for subcollection group may need to be created
        #   See: https://firebase.google.com/docs/firestore/query-data/queries#top_of_page

        def on_snapshot(snapshots, changes, timestamp):
            for doc in snapshots:
                # doc: DocumentSnapshot = snapshots[0]
                data = doc.to_dict()
                data = {
                    self.view_model_cls.get_schema_cls().g(key): val
                    for key, val in data.items()
                }

                # ie. doc from /UserViewDAV/user_id_a/_PATCH_UserViewDAV/patch_id_1
                parent_view_ref: DocumentReference = doc.reference.parent.parent
                obj = self.instances[parent_view_ref._document_path]

                self.mutation_cls.mutate_patch_one(obj=obj, data=data)

        watch = Watch.for_query(
            query=self._get_patch_query(),
            snapshot_callback=on_snapshot,
            snapshot_class_instance=DocumentSnapshot,
            reference_class_instance=DocumentReference)

    def start(self):
        """ Generates view models and listen to changes proposed to
                the view model.

        """
        self.instances = self.generate_entries()
        # time.sleep(3)  # TODO: delete after implementing sync
        self._listen_to_patch()

    def generate_entries(self):
        raise NotImplementedError



import threading
from math import inf
from collections import defaultdict
import bisect


class SnapshotContainer:
    """
    Stores all business properties of an instance of view model

    For now similar to a dictionary, but might change to pickledb
        for debugging and testing the code.

    """

    def __init__(self):
        self.d = defaultdict(list)
        self.store = dict()
        self.lock = threading.Lock()
        self._read_times = [(-inf, -inf)]

    # def get(self, key):
    #     return self.store[key]
    #
    # def set(self, key, val):
    #     self.store[key] = val

    def set_with_timestamp(self, key: str, val, timestamp: tuple=None) -> None:
        self.store[(timestamp, key)] = val
        self.d[key].append(timestamp)
        # Since the timestamps for all TimeMap.set operations
        #   are strictly increasing
        # self.d[key].sort()

    def set(self, key: str, val, timestamp: tuple=None) -> None:
        self.store[(timestamp, key)] = val
        self.d[key].append(timestamp)
        # Since the timestamps for all TimeMap.set operations
        #   are strictly increasing
        # self.d[key].sort()

    def has_previous(self, key: str):
        return len(self.d[key]) != 0

    def previous(self, key):
        ts = self.d[key][-1]
        return self.store[(ts, key)]

    def get(self, key: str, timestamp: tuple=None):
        if timestamp is None:
            return self.store[(timestamp, key)]

        idx = bisect.bisect_right(self.d[key], timestamp)
        if idx == 0:
            raise AttributeError
        else:
            ts = self.d[key][idx - 1]
            return self.store[(ts, key)]

    from math import inf

    def get_with_range(self, key, lo_excl=(-inf,-inf), hi_incl=(inf,inf)):
        """ NOTE: low is exclusive and high is inclusive (different from range)

        :param key:
        :param lo_excl:
        :param hi_incl:
        :return:
        """
        start_idx = bisect.bisect_right(self.d[key], lo_excl)
        end_idx = bisect.bisect_right(self.d[key], hi_incl)
        for ts in self.d[key][start_idx:end_idx]:
            yield self.store[(ts, key)]



from collections import namedtuple
from datetime import datetime

from onto.domain_model import DomainModel
from onto.models.base import Serializable
from onto import attrs


class UserTime(Serializable):
    time: datetime = attrs.bproperty()
    timestamp = attrs.integer(import_enabled=False)

    @timestamp.getter
    def timestamp(self):
        return self.time.timestamp()


class Location(DomainModel):

    latitude = attrs.bproperty(type_cls=float)
    longitude = attrs.bproperty(type_cls=float)


class RideFailureEvent(DomainModel):

    ride = attrs.relation(dm_cls='RideOrder')


class RideOrder(DomainModel):

    status = attrs.string()
    category = attrs.string()
    create_time = attrs.embed(obj_cls=UserTime)
    start_location = attrs.relation(dm_cls=Location)
    group = attrs.relation(dm_cls='Group')

    def add_group(self, group):
        if not self.has_group():
            self.group = group
        else:
            raise ValueError

    has_group = attrs.bproperty(type_cls=bool)

    @has_group.getter
    def has_group(self):
        return self.group is not None

    def fail_ride(self):
        failure_event = RideFailureEvent.new(ride=self)
        failure_event.save()
        self.status = 'open'


class Group(DomainModel):
    rides = attrs.relation(dm_cls='RideOrder', collection=set)

    def add_ride(self, ride):
        if ride not in self.rides:
            self.rides.add(ride)
        else:
            raise ValueError


class FailureMessage(DomainModel):

    text = attrs.bproperty(type_cls=str)


class JoinRideDirective(DomainModel):
    status = attrs.bproperty(type_cls=str)
    group = attrs.relation(dm_cls='Group')
    ride = attrs.relation(dm_cls='RideOrder')
    exception = attrs.embed()
    message = attrs.relation(dm_cls=FailureMessage, collection=list)

    def execute_new(self):
        try:
            self.group.add_ride(self.ride)
            self.ride.add_group(self.group)
        except:
            self.message = FailureMessage.new(text="Failure")
            self.status = "failed"


class RideTime(DomainModel):

    order = attrs.relation(dm_cls=RideOrder)
    category = attrs.string(import_enabled=False)
    timestamp = attrs.bproperty(type_cls=float)

    @category.getter
    def category(self):
        return self.order.category

    @timestamp.getter
    def timestamp(self):
        return self.order.time.timestamp



# def test_table():
#     from sqlalchemy import MetaData, Table, Column, Integer, String, \
#         create_engine, select
#
#     metadata = MetaData()
#
#     user = Table('user', metadata,
#                  Column('user_id', Integer, primary_key=True),
#                  Column('user_name', String(16), nullable=False),
#                  Column('email_address', String(60)),
#                  Column('nickname', String(50), nullable=False)
#     )
#     user.named_with_column = True
#
#     engine = create_engine('sqlite:///:memory:')
#     metadata.create_all(engine)
#
#     with engine.connect() as con:
#
#         k = user.insert().values({
#             'user_id': 1,
#             'user_name': 'Bill',
#             'email_address': 'ss@sss.com',
#             'nickname': 'billy'
#         })
#
#         con.execute(k)
#
#         i = user.select()
#         i = con.execute(i)
#         for row in i:
#             # assert row == (1, 'Bill', 'ss@sss.com', 'billy')
#             print(dict(row))
#             # from sqlalchemy.ext.serializer import loads
#             # res = loads(row., metadata=metadata, scoped_session=con, engine=engine)
#             # print(res)
#
#
# # def test_create_table():
# #
# #     from pyflink.dataset import ExecutionEnvironment
# #     from pyflink.table import TableConfig, DataTypes, StreamTableEnvironment
# #     from pyflink.table.descriptors import
# #
# #     exec_env = ExecutionEnvironment.get_execution_environment()
# #     exec_env.set_parallelism(1)
# #     t_config = TableConfig()
# #     t_env = StreamTableEnvironment.create(exec_env, t_config)
# #     t_env.connect(.)
# #     t_env.execute_sql(
# #         "CREATE TABLE Orders (`user` BIGINT, product STRING, amount INT) WITH (...)")
# #
#
#
# def test_couchdb():
#     from onto.context import Context
#
#     # import couchdb
#     # couch: couchdb.Server = Context.dbs.couch
#     # couch.create('mydb')
#     # db = couch['mydb']
#     # doc = {'hello': 'world'}
#     # db.save(doc)
#
#     from onto.database.couch import CouchDatabase, Reference, Snapshot
#     db = CouchDatabase()
#     ref = Reference().child('users').child('userid1').child('bookings').child('bookingid1')
#     snapshot = Snapshot(hello='world')
#     db.set(
#         ref=ref,
#         snapshot=snapshot
#     )
#
#
# def test_couchbase():
#     from onto.context import Context
#
#     # import couchdb
#     # couch: couchdb.Server = Context.dbs.couch
#     # couch.create('mydb')
#     # db = couch['mydb']
#     # doc = {'hello': 'world'}
#     # db.save(doc)
#
#     from onto.database.couchbase import CouchbaseDatabase, Reference, Snapshot
#     db = CouchbaseDatabase()
#     ref = Reference().child('users').child('userid1').child('bookings').child('bookingid1')
#     snapshot = Snapshot(hello='world')
#     db.set(
#         ref=ref,
#         snapshot=snapshot
#     )



import time

from google.cloud.firestore import DocumentReference

from examples.luggage_models import LuggageItem, Luggages
from onto import fields, testing_utils
from onto.mapper.schema import Schema
from onto.domain_model import DomainModel
from onto.view_model import ViewModel
from onto.collection_mixin import CollectionMixin
from onto.models.factory import ClsFactory

import pytest

from .fixtures import CTX


@pytest.mark.usefixtures("CTX")
def test_new(CTX):

    class EmptySchema(Schema):
        pass

    class DomainObjectBase(DomainModel):
        _collection_name = "tst_domain_objs"

    class DomainSchema1(Schema):
        property_a = fields.Raw(load_from="propertyA", dump_to="propertyA")

    DomainObject1 = ClsFactory.create(
        name="DomainObject1",
        schema=DomainSchema1,
        base=DomainObjectBase
    )

    class DomainSchema2(Schema):
        property_b = fields.Raw(load_from="propertyB", dump_to="propertyB")

    DomainObject2 = ClsFactory.create(
        name="DomainObject2",
        schema=DomainSchema2,
        base=DomainObjectBase
    )

    dependent_obj_1 = DomainObject1.new("dependentId1")
    ref_1_str: str = dependent_obj_1.doc_ref.path
    assert ref_1_str == "tst_domain_objs/dependentId1"

    dependent_obj_2 = DomainObject2.new("dependentId2")
    ref_2_str: str = dependent_obj_2.doc_ref.path
    assert ref_2_str == "tst_domain_objs/dependentId2"

    # TODO: implement

    ViewObject = ClsFactory.create(
        name="TestViewObject",
        schema=EmptySchema,
        base_tuple=(ViewModel, )
    )

    # obj = TestViewObject.new("viewId1")


@pytest.mark.skip(reason="DocumentAsView with mutable struct "
                         "is no longer supported for now. ")
def test_binding(CTX):
    d_a = {
        "luggage_type": "large",
        "weight_in_lbs": 20,
        "obj_type": "LuggageItem"
    }
    id_a = "luggage_id_a"
    obj_a = LuggageItem.new(id_a)
    obj_a._import_properties(d_a)
    obj_a.save()

    d_b = {
        "luggage_type": "medium",
        "weight_in_lbs": 15,
        "obj_type": "LuggageItem"
    }
    id_b = "luggage_id_b"
    obj_b = LuggageItem.new(id_b)
    obj_b._import_properties(d_b)
    obj_b.save()

    vm_ref: DocumentReference = CTX.db.document(
        "test_lugagges/user_a_luggages")

    vm: Luggages = Luggages.new(doc_ref=vm_ref)

    vm.bind_to(key=id_a, obj_type=LuggageItem, doc_id=id_a)
    vm.bind_to(key=id_b, obj_type=LuggageItem, doc_id=id_b)
    vm.register_listener()

    # Takes time to propagate changes
    testing_utils._wait(factor=.7)

    assert vm.to_dict() == {
        "luggages": [
            {
                "luggage_type": "large",
                "weight_in_lbs": 20
            },
            {
                "luggage_type": "medium",
                "weight_in_lbs": 15
            }
        ],
        "total_weight": 35,
        "total_count": 2,
        # "obj_type": "Luggages"
    }

    # Change the weight on one of the luggages
    obj_b.weight_in_lbs = 25
    obj_b.save()

    # Note that the update is eventually consistent with delays
    #   and local copies of a ViewModel should not be used to read
    #   updated values

    testing_utils._wait(factor=.7)

    # # Test that the view model now has updated values
    assert vm.to_dict() == {
        "luggages": [
            {
                "luggage_type": "large",
                "weight_in_lbs": 20
            },
            {
                "luggage_type": "medium",
                "weight_in_lbs": 25
            }
        ],
        "total_weight": 45,
        "total_count": 2
    }



import time

from flasgger import SwaggerView
from flask import request
from flask.json import jsonify

from examples.meeting_room.view_models import MeetingSession


class ListGet(SwaggerView):

    tags = ["MeetingSession"]

    description = "List all meeting sessions "

    responses = {
            200: {
                "description": description,
                "schema": MeetingSession.get_schema_cls()
            }
        }

    def get(self, *args, **kwargs):
        query_d = request.args.to_dict()
        meeting_sessions = MeetingSession.get_many_from_query(
            query_d=query_d, once=False)
        # time.sleep(1)  # TODO: delete after implementing sync
        results = {meeting_session.meeting_id: meeting_session.to_view_dict()
                   for meeting_session in meeting_sessions}
        return jsonify({
            "results": results
        })




def test_pypika():
    from pypika import Table, MySQLQuery as Query

    # history, customers = Table('history'), Table('customers')

    #     SELECT {order_s}.oid as oid, {product_s}.productName as productName
    #     FROM {order_s} LEFT JOIN {product_s} ON {order_s}.pid = {product_s}.productKey

    Order = Table('Order')
    Product = Table('Product')

    q = Query \
        .from_(Order) \
        .left_join(Product) \
        .on(Order.pid == Product.productKey) \
        .select(Order.oid.as_('oid'), Product.productName.as_('productName'))
    s = q.get_sql()
    assert s == ('SELECT `Order`.`oid` `oid`,`Product`.`productName` `productName` FROM '
                 '`Order` LEFT JOIN `Product` ON `Order`.`pid`=`Product`.`productKey`')


    # customers = Table('customers')
    # q = Query.from_(customers).select(
    #     customers.id, customers.fname, customers.lname, customers.phone)
    # s = q.get_sql()
    # assert s == ""



from unittest.mock import patch, call, Mock

import pytest
from google.cloud.firestore_v1 import DocumentReference, Transaction
from testfixtures import compare

from onto import schema, fields, attrs
from onto.context import Context as CTX
from onto.config import Config
from onto.database import Reference, Snapshot, is_reference
from onto.domain_model import DomainModel
from onto.models.factory import ClsFactory
from onto.firestore_object import FirestoreObject
from onto.mapper.helpers import RelationshipReference
from onto.primary_object import PrimaryObject
from onto.query import run_transaction


# Creates a schema for serializing and deserializing to firestore database
from onto.primary_object import PrimaryObjectSchema
from onto.store.gallery import Gallery


class TestObjectSchema(PrimaryObjectSchema):
    # Describes how obj.int_a is read from and stored to a document in firestore
    int_a = fields.Raw(
        # load_from="intA",
        # reads obj.int_a from firestore document field "intA"
        # dump_to="intA"
        # stores the value of obj.int_a to firestore document field "intA"
    )
    int_b = fields.Raw(
        # load_from="intB", dump_to="intB"
                       )

primary_doc_ref = CTX.db.ref/'primary'/'test_document_0'

class ContainsNestedchema(schema.Schema):
    inner = fields.Relationship(nested=True)

class ContainsNested(FirestoreObject):

    inner = attrs.relation(dm_cls='ContainsNested', nested=True)

    @property
    def doc_ref(self) -> DocumentReference:
        return primary_doc_ref


# Declares the object
class TestObject(PrimaryObject):
    _schema_cls = TestObjectSchema

# TestObject = ClsFactory.new(
#     name="TestObject",
#     schema=TestObjectSchema,
#     base=PrimaryObject
# )


def test_create_obj():

    # Creates an object with default values with reference: "TestObject/testObjId1"
    #   (but not saved to database)
    obj = TestObject.new(doc_id="testObjId1")
    assert obj.doc_id == "testObjId1"
    assert obj.collection_name == "TestObject"

    # Assigns value to the newly created object
    obj.int_a = 1
    obj.int_b = 2

    # Saves to firestore "TestObject/testObjId1"
    obj.save()

    # Gets the object from firestore "TestObject/testObjId1"
    retrieved_obj = TestObject.get(doc_id="testObjId1")

    # Access values of the object retrieved
    assert retrieved_obj.int_a == 1
    assert retrieved_obj.int_b == 2

    # Deletes the object from firestore "TestObject/testObjId1"
    retrieved_obj.delete()


def test_import_iterable():

    class ContainsIter(DomainModel):
        pass

    doc_ref_1, doc_ref_2 = Reference("hello/1"), Reference("hello/2")
    res = ContainsIter._import_val(val=[doc_ref_1, doc_ref_2])
    assert res == [doc_ref_1, doc_ref_2]
    res = ContainsIter._import_val(val={1: doc_ref_1, 2: doc_ref_2})
    assert res == {1: doc_ref_1, 2: doc_ref_2}


def test_import_nested():

    inner = setup_object(doc_id="inner")

    transaction = CTX.db.transaction()
    doc_ref = inner.doc_ref
    snapshot = Snapshot(foo='bar')

    """
    Import document reference from database as object 
    """
    with patch.object(CTX.db, 'get', return_value=snapshot) as mock_method:
        store = Gallery()
        field_deserialized = RelationshipReference(
            nested=True,
            doc_ref=doc_ref,
            obj_type=TestObject
        )
        res = ContainsNested._import_val(val=field_deserialized, _store=store)

        def _get_snapshots(database, transaction, refs):
            return list(
                (ref, database.get(ref=ref))
                for ref in refs
            )

        store.refresh(transaction=None, get_snapshots=_get_snapshots)
        assert isinstance(res, TestObject)
        mock_method.assert_called_once()

    """ Import document reference from database without circular reference 
            issues. 
    """
    with patch.object(CTX.db, 'get', return_value=snapshot) as mock_method:
        obj = ContainsNested.new()
        obj.inner = obj  # An example of circularly referenced object
        obj.save()
        store = Gallery()
        d = {'obj_type': 'ContainsNested', 'doc_ref': 'primary/test_document_0', 'inner': RelationshipReference(nested=True, doc_ref=obj.doc_ref)}
        ContainsNested._import_from_dict(d=d, _store=store)
        assert isinstance(res, TestObject)

    """
    Import document reference from database as object that 
        has transaction 
    """
    with patch.object(CTX.db, 'get', return_value=snapshot) as mock_method:
        obj = ContainsNested(transaction=transaction)

        field_deserialized = RelationshipReference(
            nested=True,
            doc_ref=doc_ref,
            obj_type=ContainsNested
        )
        store = Gallery()

        res = obj._import_val(
            val=field_deserialized, _store=store)
        store.refresh(transaction=None)
        assert isinstance(res, TestObject)
        # assert res.transaction == transaction
        # mock_method.assert_called_with(
        #     transaction=transaction
        # )

    delete_object(inner.doc_id)


def test_export_nested():

    inner = setup_object(doc_id="inner")

    transaction = CTX.db.transaction()
    doc_ref = inner.doc_ref
    # snapshot = doc_ref.get()

    """
    Export value 
    """
    with patch.object(CTX.db, 'set', return_value=None) as mock_method:
        obj = ContainsNested.new(inner=inner)

        field_deserialized = RelationshipReference(
            nested=True,
            obj=obj.inner
        )
        res = obj._export_val(val=field_deserialized)
        assert isinstance(res, str)
        assert mock_method.call_args == call(snapshot=inner.to_snapshot(), ref=doc_ref, transaction=None)

    """
    Set object in relationship reference with transaction
    """
    with patch.object(CTX.db, 'set', return_value=None) as mock_method:
        obj = ContainsNested.new(inner=inner, transaction=transaction)

        field_deserialized = RelationshipReference(
            nested=True,
            obj=obj.inner
        )
        res = obj._export_val(
            val=field_deserialized, transaction=transaction)
        assert isinstance(res, str)
        assert mock_method.call_args == call(snapshot=inner.to_snapshot(), ref=doc_ref, transaction=transaction)

    # with patch.object(CTX.db, 'set', return_value=None) as mock_method:
    #     obj = ContainsNested.new(inner=inner, transaction=transaction)
    #     obj.save(transaction=transaction)
    #     assert mock_method.call_args == call(snapshot=inner.to_snapshot(), ref=doc_ref, transaction=transaction)
    #
    # with patch.object(CTX.db, 'set', return_value=None) as mock_method:
    #
    #     # global primary_doc_ref
    #     obj = ContainsNested.new(inner=inner, transaction=transaction, doc_ref=primary_doc_ref)
    #     obj._export_as_dict(transaction=transaction)
    #
    #     # do_stuffs()
    #
    #     assert mock_method.call_args == call(snapshot=inner._export_as_dict(), ref=doc_ref, transaction=transaction)

    # with patch.object(CTX.db, 'set', return_value=None) as mock_method:
    #     transaction_invoked = None
    #
    #     @run_transaction
    #     def do_stuffs(transaction: Transaction):
    #         global primary_doc_ref
    #         nonlocal transaction_invoked
    #         transaction_invoked = transaction
    #         obj = ContainsNested.new(inner=inner, transaction=transaction, doc_ref=primary_doc_ref)
    #         obj.save(transaction=transaction)
    #
    #     do_stuffs()
    #
    #     assert mock_method.call_args == call(snapshot=inner._export_as_dict(), ref=doc_ref, transaction=transaction_invoked)

    # with patch.object(CTX.db, 'set', return_value=None) as mock_method:
    #
    #     @run_transaction
    #     def do_stuffs(transaction: Transaction):
    #         global primary_doc_ref
    #         # transaction.set = Mock(return_value=None)
    #         obj = ContainsNested.new(inner=inner, doc_ref=primary_doc_ref)
    #         obj._export_as_dict()
    #
    #     do_stuffs()
    #
    #     assert mock_method.call_args == call(snapshot=inner._export_as_dict(), ref=doc_ref, transaction=transaction)

    """
    Tests that transactions are implicitly passed to referenced objects 
    """

    with patch.object(CTX.db, 'set', return_value=None) as mock_method:

        transaction_invoked = None

        @run_transaction
        def do_stuffs(transaction: Transaction):
            nonlocal transaction_invoked
            transaction_invoked = transaction

            # transaction.set = Mock(return_value=None)
            global primary_doc_ref
            obj = ContainsNested.new(inner=inner, doc_ref=primary_doc_ref)
            obj.save()

        do_stuffs()

        assert mock_method.call_args_list == [
            call(snapshot=inner._export_as_dict(), ref=doc_ref, transaction=transaction_invoked),
            call(snapshot={'obj_type': 'ContainsNested', 'inner': doc_ref}, ref=primary_doc_ref, transaction=transaction_invoked)
        ]

    delete_object(inner.doc_id)


def test_relationship_not_nested():
    referenced_obj = TestObject.new(doc_id="testObjId1")

    # Assigns value to the newly created object
    referenced_obj.int_a = 1
    referenced_obj.int_b = 2

    # Saves to firestore "TestObject/testObjId1"
    referenced_obj.save()

    from onto.primary_object import PrimaryObjectSchema
    class MasterObjectSchema(PrimaryObjectSchema):
        # Describes how obj.int_a is read from and stored to a document in firestore
        nested_ref = fields.Relationship(nested=False)

    MasterObject = ClsFactory.create(
        name="MasterObject",
        schema=MasterObjectSchema,
        base=PrimaryObject
    )

    master_obj = MasterObject.new(doc_id="masterObjId1")
    master_obj.nested_ref = referenced_obj.doc_ref
    assert master_obj._export_as_dict() == {
        'doc_ref': 'MasterObject/masterObjId1',
        'doc_id': 'masterObjId1',
        'obj_type': 'MasterObject',
        "nestedRef": CTX.db.ref/"TestObject"/"testObjId1"
    }

    master_obj.delete()
    CTX.db.delete(ref=CTX.db.ref/"TestObject"/"testObjId1")


def test_relationship_nested():
    referenced_obj = TestObject.new(doc_id="testObjId3")

    # Assigns value to the newly created object
    referenced_obj.int_a = 1
    referenced_obj.int_b = 2

    from onto.primary_object import PrimaryObjectSchema
    class MasterObjectSchemaNested(PrimaryObjectSchema):        # Describes how obj.int_a is read from and stored to a document in firestore
        nested_obj = fields.Relationship(nested=True)

    MasterObjectNested = ClsFactory.create(
        name="MasterObjectNested",
        schema=MasterObjectSchemaNested,
        base=PrimaryObject
    )

    master_obj = MasterObjectNested.new(doc_id="masterObjIdNested1")
    master_obj.nested_obj = referenced_obj

    assert master_obj._export_as_dict() == {
        'doc_id': 'masterObjIdNested1',
        'doc_ref': 'MasterObjectNested/masterObjIdNested1',
        'obj_type': 'MasterObjectNested',
        "nestedObj": referenced_obj.doc_ref
    }

    master_obj.save()

    doc_ref = CTX.db.ref/"TestObject"/"testObjId3"
    assert CTX.db.get(ref=doc_ref).to_dict() == \
           {
               "intA": 1,
               "intB": 2,
               "obj_type": "TestObject",
               'doc_id': 'testObjId3',
               'doc_ref': "TestObject/testObjId3"
           }

    master_obj.delete()
    CTX.db.delete(ref=doc_ref)


def setup_object(doc_id):

    # Creates an object with default values with reference: "TestObject/testObjId1"
    #   (but not saved to database)
    obj = TestObject.new(doc_id=doc_id)
    assert obj.doc_id == doc_id
    assert obj.collection_name == "TestObject"

    # Assigns value to the newly created object
    obj.int_a = 1
    obj.int_b = 2

    # Saves to firestore "TestObject/testObjId1"
    obj.save()

    return obj


def delete_object(doc_id):

    # Creates an object with default values with reference: "TestObject/testObjId1"
    #   (but not saved to database)
    obj = TestObject.new(doc_id=doc_id)

    # Deletes "TestObject/testObjId1"
    obj.delete()


def test_stream_objects():

    setup_object(doc_id="testObjId1")
    setup_object(doc_id="testObjId2")

    for obj in TestObject.all():
        assert obj.doc_id in {"testObjId1", "testObjId2"}
        assert obj.to_dict().items() >= {
            "intA": 1,
            "intB": 2,
            "obj_type": "TestObject"
        }.items()

    for obj in TestObject.all():
        obj.delete()



from onto.domain_model import DomainModel
from onto import attrs


class City(DomainModel):

    city_name = attrs.bproperty()
    country = attrs.bproperty()
    capital = attrs.bproperty()

    class Meta:
        collection_name = "City"


class Municipality(City):
    pass


class StandardCity(City):
    city_state = attrs.bproperty()
    regions = attrs.bproperty()



from .base import Sink


class Websocket(Sink):

    def __init__(self, namespace):
        self._namespace = namespace

    def emit(self, obj):
        self._namespace.emit("updated", obj.to_view_dict())

websocket = Websocket



from itertools import count

import pytest as pytest

from onto import schema, fields, view_model

from onto.store.struct import Struct
from tests.color_fixtures import Color, PaletteViewModel, RainbowStoreBpss, color_refs, vm
from .fixtures import CTX


@pytest.fixture
def v_cls(CTX):

    class RainbowSchema(schema.Schema):
        rainbow_name = fields.Raw(dump_only=True)
        colors = fields.Raw(dump_only=True)

    class RainbowView(view_model.ViewModel):

        _schema_cls = RainbowSchema

        _color_d = dict()

        def __init__(self, order_d=None, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.order = order_d

        @property
        def colors(self):
            return [self._color_d[key] for key in sorted(self._color_d)]

        @property
        def rainbow_name(self):
            return "-".join([self._color_d[key] for key in sorted(self._color_d)])

        def set_color(self, color_name, order):
            self._color_d[(order, color_name)] = color_name

        def get_vm_update_callback(self, dm_cls, *args, **kwargs):
            if issubclass(dm_cls, Color):
                def update_func(vm: RainbowView, dm: Color):
                    order = self.order[dm.doc_id]
                    vm.set_color(dm.name, order=order)
                return update_func

        @classmethod
        def get_from_color_names(cls, color_names):

            struct = Struct(schema_obj=RainbowStoreBpss())

            order_gen = count()
            order_d = dict()

            for color_name in color_names:
                obj_type = Color
                doc_id = "doc_id_{}".format(color_name)

                order_d[doc_id] = next(order_gen)

                struct["colors"][doc_id] = (obj_type, doc_id,)

            return cls.get(struct_d=struct, order_d=order_d, once=True)

    return RainbowView


def test_to_dict_view(v_cls, color_refs):
    vm = v_cls.get_from_color_names(["yellow", "magenta", "cian"])
    # time.sleep(3)
    assert vm.to_view_dict() == {
        'rainbowName': 'yellow-magenta-cian',
        'colors': ['yellow', 'magenta', 'cian']
    }


def test_vm__export_as_view_dict(color_refs, CTX):
    vm = PaletteViewModel.new(
        doc_ref=CTX.db.collection("palettes").document("palette_id_a")
    )
    vm.palette_name = 'cmyk'
    vm.colors = color_refs
    assert vm.to_dict() == {
        "paletteName": 'cmyk',
        "colors": [
            {'name': 'cian'},
            {'name': 'magenta'},
            {'name': 'yellow'},
            {'name': 'black'}
        ]
    }


def test_vm(vm: PaletteViewModel):
    assert vm.to_dict() == {
        "paletteName": 'cmyk',
        "colors": [
            {'name': 'cian'},
            {'name': 'magenta'},
            {'name': 'yellow'},
            {'name': 'black'}
        ]
    }



"""
Ref: http://code.activestate.com/recipes/442447/

META:LICENSE=PSF

OpenSource under PSF and http://creativecommons.org/licenses/by-sa/3.0/
Downloaded as a snippet from activestate.com

http://code.activestate.com/help/terms/

Attribution-ShareAlike 3.0 Unported

CREATIVE COMMONS CORPORATION IS NOT A LAW FIRM AND DOES NOT PROVIDE LEGAL SERVICES. DISTRIBUTION OF THIS LICENSE DOES NOT CREATE AN ATTORNEY-CLIENT RELATIONSHIP. CREATIVE COMMONS PROVIDES THIS INFORMATION ON AN "AS-IS" BASIS. CREATIVE COMMONS MAKES NO WARRANTIES REGARDING THE INFORMATION PROVIDED, AND DISCLAIMS LIABILITY FOR DAMAGES RESULTING FROM ITS USE.
License

THE WORK (AS DEFINED BELOW) IS PROVIDED UNDER THE TERMS OF THIS CREATIVE COMMONS PUBLIC LICENSE ("CCPL" OR "LICENSE"). THE WORK IS PROTECTED BY COPYRIGHT AND/OR OTHER APPLICABLE LAW. ANY USE OF THE WORK OTHER THAN AS AUTHORIZED UNDER THIS LICENSE OR COPYRIGHT LAW IS PROHIBITED.

BY EXERCISING ANY RIGHTS TO THE WORK PROVIDED HERE, YOU ACCEPT AND AGREE TO BE BOUND BY THE TERMS OF THIS LICENSE. TO THE EXTENT THIS LICENSE MAY BE CONSIDERED TO BE A CONTRACT, THE LICENSOR GRANTS YOU THE RIGHTS CONTAINED HERE IN CONSIDERATION OF YOUR ACCEPTANCE OF SUCH TERMS AND CONDITIONS.

1. Definitions

"Adaptation" means a work based upon the Work, or upon the Work and other pre-existing works, such as a translation, adaptation, derivative work, arrangement of music or other alterations of a literary or artistic work, or phonogram or performance and includes cinematographic adaptations or any other form in which the Work may be recast, transformed, or adapted including in any form recognizably derived from the original, except that a work that constitutes a Collection will not be considered an Adaptation for the purpose of this License. For the avoidance of doubt, where the Work is a musical work, performance or phonogram, the synchronization of the Work in timed-relation with a moving image ("synching") will be considered an Adaptation for the purpose of this License.
"Collection" means a collection of literary or artistic works, such as encyclopedias and anthologies, or performances, phonograms or broadcasts, or other works or subject matter other than works listed in Section 1(f) below, which, by reason of the selection and arrangement of their contents, constitute intellectual creations, in which the Work is included in its entirety in unmodified form along with one or more other contributions, each constituting separate and independent works in themselves, which together are assembled into a collective whole. A work that constitutes a Collection will not be considered an Adaptation (as defined below) for the purposes of this License.
"Creative Commons Compatible License" means a license that is listed at https://creativecommons.org/compatiblelicenses that has been approved by Creative Commons as being essentially equivalent to this License, including, at a minimum, because that license: (i) contains terms that have the same purpose, meaning and effect as the License Elements of this License; and, (ii) explicitly permits the relicensing of adaptations of works made available under that license under this License or a Creative Commons jurisdiction license with the same License Elements as this License.
"Distribute" means to make available to the public the original and copies of the Work or Adaptation, as appropriate, through sale or other transfer of ownership.
"License Elements" means the following high-level license attributes as selected by Licensor and indicated in the title of this License: Attribution, ShareAlike.
"Licensor" means the individual, individuals, entity or entities that offer(s) the Work under the terms of this License.
"Original Author" means, in the case of a literary or artistic work, the individual, individuals, entity or entities who created the Work or if no individual or entity can be identified, the publisher; and in addition (i) in the case of a performance the actors, singers, musicians, dancers, and other persons who act, sing, deliver, declaim, play in, interpret or otherwise perform literary or artistic works or expressions of folklore; (ii) in the case of a phonogram the producer being the person or legal entity who first fixes the sounds of a performance or other sounds; and, (iii) in the case of broadcasts, the organization that transmits the broadcast.
"Work" means the literary and/or artistic work offered under the terms of this License including without limitation any production in the literary, scientific and artistic domain, whatever may be the mode or form of its expression including digital form, such as a book, pamphlet and other writing; a lecture, address, sermon or other work of the same nature; a dramatic or dramatico-musical work; a choreographic work or entertainment in dumb show; a musical composition with or without words; a cinematographic work to which are assimilated works expressed by a process analogous to cinematography; a work of drawing, painting, architecture, sculpture, engraving or lithography; a photographic work to which are assimilated works expressed by a process analogous to photography; a work of applied art; an illustration, map, plan, sketch or three-dimensional work relative to geography, topography, architecture or science; a performance; a broadcast; a phonogram; a compilation of data to the extent it is protected as a copyrightable work; or a work performed by a variety or circus performer to the extent it is not otherwise considered a literary or artistic work.
"You" means an individual or entity exercising rights under this License who has not previously violated the terms of this License with respect to the Work, or who has received express permission from the Licensor to exercise rights under this License despite a previous violation.
"Publicly Perform" means to perform public recitations of the Work and to communicate to the public those public recitations, by any means or process, including by wire or wireless means or public digital performances; to make available to the public Works in such a way that members of the public may access these Works from a place and at a place individually chosen by them; to perform the Work to the public by any means or process and the communication to the public of the performances of the Work, including by public digital performance; to broadcast and rebroadcast the Work by any means including signs, sounds or images.
"Reproduce" means to make copies of the Work by any means including without limitation by sound or visual recordings and the right of fixation and reproducing fixations of the Work, including storage of a protected performance or phonogram in digital form or other electronic medium.
2. Fair Dealing Rights. Nothing in this License is intended to reduce, limit, or restrict any uses free from copyright or rights arising from limitations or exceptions that are provided for in connection with the copyright protection under copyright law or other applicable laws.

3. License Grant. Subject to the terms and conditions of this License, Licensor hereby grants You a worldwide, royalty-free, non-exclusive, perpetual (for the duration of the applicable copyright) license to exercise the rights in the Work as stated below:

to Reproduce the Work, to incorporate the Work into one or more Collections, and to Reproduce the Work as incorporated in the Collections;
to create and Reproduce Adaptations provided that any such Adaptation, including any translation in any medium, takes reasonable steps to clearly label, demarcate or otherwise identify that changes were made to the original Work. For example, a translation could be marked "The original work was translated from English to Spanish," or a modification could indicate "The original work has been modified.";
to Distribute and Publicly Perform the Work including as incorporated in Collections; and,
to Distribute and Publicly Perform Adaptations.
For the avoidance of doubt:

Non-waivable Compulsory License Schemes. In those jurisdictions in which the right to collect royalties through any statutory or compulsory licensing scheme cannot be waived, the Licensor reserves the exclusive right to collect such royalties for any exercise by You of the rights granted under this License;
Waivable Compulsory License Schemes. In those jurisdictions in which the right to collect royalties through any statutory or compulsory licensing scheme can be waived, the Licensor waives the exclusive right to collect such royalties for any exercise by You of the rights granted under this License; and,
Voluntary License Schemes. The Licensor waives the right to collect royalties, whether individually or, in the event that the Licensor is a member of a collecting society that administers voluntary licensing schemes, via that society, from any exercise by You of the rights granted under this License.
The above rights may be exercised in all media and formats whether now known or hereafter devised. The above rights include the right to make such modifications as are technically necessary to exercise the rights in other media and formats. Subject to Section 8(f), all rights not expressly granted by Licensor are hereby reserved.

4. Restrictions. The license granted in Section 3 above is expressly made subject to and limited by the following restrictions:

You may Distribute or Publicly Perform the Work only under the terms of this License. You must include a copy of, or the Uniform Resource Identifier (URI) for, this License with every copy of the Work You Distribute or Publicly Perform. You may not offer or impose any terms on the Work that restrict the terms of this License or the ability of the recipient of the Work to exercise the rights granted to that recipient under the terms of the License. You may not sublicense the Work. You must keep intact all notices that refer to this License and to the disclaimer of warranties with every copy of the Work You Distribute or Publicly Perform. When You Distribute or Publicly Perform the Work, You may not impose any effective technological measures on the Work that restrict the ability of a recipient of the Work from You to exercise the rights granted to that recipient under the terms of the License. This Section 4(a) applies to the Work as incorporated in a Collection, but this does not require the Collection apart from the Work itself to be made subject to the terms of this License. If You create a Collection, upon notice from any Licensor You must, to the extent practicable, remove from the Collection any credit as required by Section 4(c), as requested. If You create an Adaptation, upon notice from any Licensor You must, to the extent practicable, remove from the Adaptation any credit as required by Section 4(c), as requested.
You may Distribute or Publicly Perform an Adaptation only under the terms of: (i) this License; (ii) a later version of this License with the same License Elements as this License; (iii) a Creative Commons jurisdiction license (either this or a later license version) that contains the same License Elements as this License (e.g., Attribution-ShareAlike 3.0 US)); (iv) a Creative Commons Compatible License. If you license the Adaptation under one of the licenses mentioned in (iv), you must comply with the terms of that license. If you license the Adaptation under the terms of any of the licenses mentioned in (i), (ii) or (iii) (the "Applicable License"), you must comply with the terms of the Applicable License generally and the following provisions: (I) You must include a copy of, or the URI for, the Applicable License with every copy of each Adaptation You Distribute or Publicly Perform; (II) You may not offer or impose any terms on the Adaptation that restrict the terms of the Applicable License or the ability of the recipient of the Adaptation to exercise the rights granted to that recipient under the terms of the Applicable License; (III) You must keep intact all notices that refer to the Applicable License and to the disclaimer of warranties with every copy of the Work as included in the Adaptation You Distribute or Publicly Perform; (IV) when You Distribute or Publicly Perform the Adaptation, You may not impose any effective technological measures on the Adaptation that restrict the ability of a recipient of the Adaptation from You to exercise the rights granted to that recipient under the terms of the Applicable License. This Section 4(b) applies to the Adaptation as incorporated in a Collection, but this does not require the Collection apart from the Adaptation itself to be made subject to the terms of the Applicable License.
If You Distribute, or Publicly Perform the Work or any Adaptations or Collections, You must, unless a request has been made pursuant to Section 4(a), keep intact all copyright notices for the Work and provide, reasonable to the medium or means You are utilizing: (i) the name of the Original Author (or pseudonym, if applicable) if supplied, and/or if the Original Author and/or Licensor designate another party or parties (e.g., a sponsor institute, publishing entity, journal) for attribution ("Attribution Parties") in Licensor's copyright notice, terms of service or by other reasonable means, the name of such party or parties; (ii) the title of the Work if supplied; (iii) to the extent reasonably practicable, the URI, if any, that Licensor specifies to be associated with the Work, unless such URI does not refer to the copyright notice or licensing information for the Work; and (iv) , consistent with Ssection 3(b), in the case of an Adaptation, a credit identifying the use of the Work in the Adaptation (e.g., "French translation of the Work by Original Author," or "Screenplay based on original Work by Original Author"). The credit required by this Section 4(c) may be implemented in any reasonable manner; provided, however, that in the case of a Adaptation or Collection, at a minimum such credit will appear, if a credit for all contributing authors of the Adaptation or Collection appears, then as part of these credits and in a manner at least as prominent as the credits for the other contributing authors. For the avoidance of doubt, You may only use the credit required by this Section for the purpose of attribution in the manner set out above and, by exercising Your rights under this License, You may not implicitly or explicitly assert or imply any connection with, sponsorship or endorsement by the Original Author, Licensor and/or Attribution Parties, as appropriate, of You or Your use of the Work, without the separate, express prior written permission of the Original Author, Licensor and/or Attribution Parties.
Except as otherwise agreed in writing by the Licensor or as may be otherwise permitted by applicable law, if You Reproduce, Distribute or Publicly Perform the Work either by itself or as part of any Adaptations or Collections, You must not distort, mutilate, modify or take other derogatory action in relation to the Work which would be prejudicial to the Original Author's honor or reputation. Licensor agrees that in those jurisdictions (e.g. Japan), in which any exercise of the right granted in Section 3(b) of this License (the right to make Adaptations) would be deemed to be a distortion, mutilation, modification or other derogatory action prejudicial to the Original Author's honor and reputation, the Licensor will waive or not assert, as appropriate, this Section, to the fullest extent permitted by the applicable national law, to enable You to reasonably exercise Your right under Section 3(b) of this License (right to make Adaptations) but not otherwise.
5. Representations, Warranties and Disclaimer

UNLESS OTHERWISE MUTUALLY AGREED TO BY THE PARTIES IN WRITING, LICENSOR OFFERS THE WORK AS-IS AND MAKES NO REPRESENTATIONS OR WARRANTIES OF ANY KIND CONCERNING THE WORK, EXPRESS, IMPLIED, STATUTORY OR OTHERWISE, INCLUDING, WITHOUT LIMITATION, WARRANTIES OF TITLE, MERCHANTIBILITY, FITNESS FOR A PARTICULAR PURPOSE, NONINFRINGEMENT, OR THE ABSENCE OF LATENT OR OTHER DEFECTS, ACCURACY, OR THE PRESENCE OF ABSENCE OF ERRORS, WHETHER OR NOT DISCOVERABLE. SOME JURISDICTIONS DO NOT ALLOW THE EXCLUSION OF IMPLIED WARRANTIES, SO SUCH EXCLUSION MAY NOT APPLY TO YOU.

6. Limitation on Liability. EXCEPT TO THE EXTENT REQUIRED BY APPLICABLE LAW, IN NO EVENT WILL LICENSOR BE LIABLE TO YOU ON ANY LEGAL THEORY FOR ANY SPECIAL, INCIDENTAL, CONSEQUENTIAL, PUNITIVE OR EXEMPLARY DAMAGES ARISING OUT OF THIS LICENSE OR THE USE OF THE WORK, EVEN IF LICENSOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

7. Termination

This License and the rights granted hereunder will terminate automatically upon any breach by You of the terms of this License. Individuals or entities who have received Adaptations or Collections from You under this License, however, will not have their licenses terminated provided such individuals or entities remain in full compliance with those licenses. Sections 1, 2, 5, 6, 7, and 8 will survive any termination of this License.
Subject to the above terms and conditions, the license granted here is perpetual (for the duration of the applicable copyright in the Work). Notwithstanding the above, Licensor reserves the right to release the Work under different license terms or to stop distributing the Work at any time; provided, however that any such election will not serve to withdraw this License (or any other license that has been, or is required to be, granted under the terms of this License), and this License will continue in full force and effect unless terminated as stated above.
8. Miscellaneous

Each time You Distribute or Publicly Perform the Work or a Collection, the Licensor offers to the recipient a license to the Work on the same terms and conditions as the license granted to You under this License.
Each time You Distribute or Publicly Perform an Adaptation, Licensor offers to the recipient a license to the original Work on the same terms and conditions as the license granted to You under this License.
If any provision of this License is invalid or unenforceable under applicable law, it shall not affect the validity or enforceability of the remainder of the terms of this License, and without further action by the parties to this agreement, such provision shall be reformed to the minimum extent necessary to make such provision valid and enforceable.
No term or provision of this License shall be deemed waived and no breach consented to unless such waiver or consent shall be in writing and signed by the party to be charged with such waiver or consent.
This License constitutes the entire agreement between the parties with respect to the Work licensed here. There are no understandings, agreements or representations with respect to the Work not specified here. Licensor shall not be bound by any additional provisions that may appear in any communication from You. This License may not be modified without the mutual written agreement of the Licensor and You.
The rights granted under, and the subject matter referenced, in this License were drafted utilizing the terminology of the Berne Convention for the Protection of Literary and Artistic Works (as amended on September 28, 1979), the Rome Convention of 1961, the WIPO Copyright Treaty of 1996, the WIPO Performances and Phonograms Treaty of 1996 and the Universal Copyright Convention (as revised on July 24, 1971). These rights and subject matter take effect in the relevant jurisdiction in which the License terms are sought to be enforced according to the corresponding provisions of the implementation of those treaty provisions in the applicable national law. If the standard suite of rights granted under applicable copyright law includes additional rights not granted under this License, such additional rights are deemed to be included in the License; this License is not intended to restrict the license of any rights under applicable law.
Creative Commons Notice

Creative Commons is not a party to this License, and makes no warranty whatsoever in connection with the Work. Creative Commons will not be liable to You or any party on any legal theory for any damages whatsoever, including without limitation any general, special, incidental or consequential damages arising in connection to this license. Notwithstanding the foregoing two (2) sentences, if Creative Commons has expressly identified itself as the Licensor hereunder, it shall have all rights and obligations of Licensor.

Except for the limited purpose of indicating to the public that the Work is licensed under the CCPL, Creative Commons does not authorize the use by either party of the trademark "Creative Commons" or any related trademark or logo of Creative Commons without the prior written consent of Creative Commons. Any permitted use will be in compliance with Creative Commons' then-current trademark usage guidelines, as may be published on its website or otherwise made available upon request from time to time. For the avoidance of doubt, this trademark restriction does not form part of the License.

Creative Commons may be contacted at https://creativecommons.org/.

"""


"""A wrapper around DBAPI-compliant databases to support iteration
and generator expression syntax for requests, instead of SQL

To get an iterator, initialize a connection to the database, then
set the cursor attribute of the query class to its cursor

Create an instance of Table for the tables you want to use

Then you can use the class query. You create an instance by passing
a generator expression as parameter. This instance translates the
generator expression in an SQL statement ; then you can iterate
on it to get the selected items as objects, dictionaries or lists

Supposing you call this module db_iterator.py, here is an example 
of use with sqlite :

    from pysqlite2 import dbapi2 as sqlite
    from db_iterator import query, Table

    conn = sqlite.connect('planes')
    query.cursor = conn.cursor()

    plane = Table()
    countries = Table()

    # all the items produced by iteration on query() are instances
    # of the Record class
    
    # simple requests
    # since no attribute of r is specified in the query, returns a list
    # of instances of Record with attributes matching all the field names
    print [ r.name for r in query(r for r in plane if r.country == 'France') ]
    
    # this request returns a list instances of Record with the attribute 
    # c_country (c.country with the . replaced by _)
    print [ country for country in query(c.country for c in countries 
            if c.continent == 'Europe') ]

    # request on two tables
    print [r.name for r in query (r for r in plane for c in countries 
            if r.country == c.country and c.continent == 'Europe')]

"""
import tokenize
import token
import types

class ge_visitor:
    """Instances of ge_visitor are used as the visitor argument to 
    compiler.walk(tree,visitor) where tree is an AST tree built by
    compiler.parse
    The instance has a src attribute which looks like the source
    code from which the tree was built
    Only a few of the visitNodeType are implemented, those likely to appear
    in a database query. Can be easily extended
    """

    def __init__(self):
        self.src = ''

    def visitTuple(self,t):
        self.src += ','.join ( [ get_source(n) for n in t.nodes ])

    def visitList(self,t):
        self.src += ','.join ( [ get_source(n) for n in t.nodes ])

    def visitMul(self,t):
        self.src += '(%s)' %('*'.join([ get_source(n) for n in t]))

    def visitName(self,t):
        self.src += t.name

    def visitConst(self,t):
        if type(t.value) is str:
            # convert single quotes, SQL-style
            self.src += "'%s'" %t.value.replace("'","''")
        else:
            self.src += str(t.value)

    def visitAssName(self,t):
        self.src += t.name

    def visitGetattr(self,t):
        self.src += '%s.%s' %(get_source(t.expr),str(t.attrname))

    def visitGenExprFor(self,t):
        self.src += 'for %s in %s ' %(get_source(t.assign),
                get_source(t.iter))
        if t.ifs:
            self.src += ' if ' +''.join([ get_source(i) for i in t.ifs ])

    def visitGenExprIf(self,t):
        self.src += get_source(t.test)

    def visitCompare(self,t):
        compiler.walk(t.expr,self)
        self.src += ' '
        for o in t.ops:
            oper = o[0]
            if oper == '==':
                oper = '='
            self.src += oper + ' '
            compiler.walk(o[1],self)

    def visitAnd(self,t):
        self.src += '('
        self.src += ' AND '.join([ get_source(n) for n in t.nodes ])
        self.src+= ')'

    def visitOr(self,t):
        self.src += '('
        self.src += ' OR '.join([ get_source(n) for n in t.nodes ])
        self.src+= ')'

    def visitNot(self,t):
        self.src += '(NOT ' + get_source(t.expr) + ')'

def get_source(node):
    """Return the source code of the node, built by an instance of
    ge_visitor"""
    return compiler.walk(node,ge_visitor()).src

class genExprVisitor:
    """Visitor used to initialize GeneratorExpression objects
    Uses the visitor pattern. See the compiler.visitor module"""

    def __init__(self):
        self.GenExprs = []

    def visitGenExprInner(self,node):
        ge = GeneratorExpression()
        self.GenExprs.append(ge)
        for y in node.getChildren():
            if y.__class__ is compiler.ast.GenExprFor:
                ge.exprfor.append(y)
            else:
                ge.result = y

class GeneratorExpression:
    """A class for a Generator Expression"""
    def __init__(self):
        self.result = None
        self.exprfor = []
        
class Record(object):
    """A generic class for database records"""
    pass

class Table:
    """A basic iterable class to avoid syntax errors"""
    def __iter__(self):
        return self
    
class query:
    """Class used for database queries
    Instance is created with query(ge) where ge is a generator
    expression
    The __init__ method builds the SQL select expression matching the
    generator expression
    Iteration on the instance of query yields the items found by
    the SQL select, under the form specified by return_type : an object,
    a dictionary or a list"""

    cursor = None   # to be set to the cursor of the connection
    return_type = object    # can be set to dict or list

    def __init__(self,s):
        self._iterating = False # used in next()

        # First we must get the source code of the generator expression
        # I use an ugly hack with stack frame attributes and tokenize
        # If there's a cleaner and safer way, please tell me !
        readline = open(s.gi_frame.f_code.co_filename).readline
        first_line = s.gi_frame.f_code.co_firstlineno
        flag = False
        self.source = ''    # the source code
        for t in tokenize.generate_tokens(open(s.gi_frame.f_code.co_filename).readline):
            # check all tokens until the last parenthesis is closed
            t_type,t_string,(r_start,c_start),(r_end,c_end),line = t
            t_name = token.tok_name[t_type]
            if r_start == first_line:
                if t_name == 'NAME' and t_string=="query":
                    flag = True
                    res = t_string
                    start = 0 # number of parenthesis
                    continue
            if flag:
                self.source += ' '+t_string
                if t_name == 'OP':
                        if t_string=='(':
                            start += 1
                        elif t_string == ')':
                            start -= 1
                            if start == 0:
                                break
        # when the source has been found, build an AST tree from it
        ast = compiler.parse(self.source.strip())
        # use a visitor to find the generator expression(s) in the source
        visitor = genExprVisitor()
        compiler.walk(ast,visitor)
        # if there are nested generator expressions, it's too difficult
        # to handle : raise an exception
        if len(visitor.GenExprs)>1:
            raise Exception('Invalid expression, found more ' \
                'than 1 generator expression')
        ge = visitor.GenExprs[0]
        self.sql = self.build_sql(ge)

    def build_sql(self,ge):
        """ Build the SQL select for the generator expression
        ge is an instance of GeneratorExpression
        The generator expression looks like
        (result) for x1 in table1 [ for x2 in table2] [ if condition ]
        It has 2 attributes :
        - result : an AST tree with the "result" part
        - exprfor : a list of AST trees, one for each "for ... in ..."
        """
        self.res = []
        if ge.result.__class__ is compiler.ast.Tuple:
            # more than one item in result
            self.res = ge.result.getChildren()
        else:
            self.res = [ge.result]
        results = [] # a list of strings = result part of the SQL expression
        for res in self.res:
            # a result can be a stand-alone name, or a "qualified" name,
            # with the table name first (table.field)
            if res.__class__ is compiler.ast.Name:
                results.append((res.name,None))
            elif res.__class__ is compiler.ast.Getattr:
                results.append((get_source(res.expr),res.attrname))
        self.results = results

        # "for x in y" produces an item in the dictionary recdefs :
        # recdef[x] = y
        recdefs = {}
        conditions = []
        for exprfor in ge.exprfor:
            recdefs[get_source(exprfor.assign)] = \
                get_source(exprfor.iter)
            if exprfor.ifs:
                # an AST tree for the condition
                conditions = exprfor.ifs

        # To build objects or dictionaries in the result set, we must
        # know the name of the fields in all the tables used in the
        # query. For this, make a simple select in each table and read
        # the information in cursor.description
        self.names={}
        for rec,table in recdefs.iteritems():
            self.cursor.execute('SELECT * FROM %s' %table)
            self.names[rec] = [ d[0] for d in self.cursor.description ]

        sql_res = [] # the way the field will appear in the SQL string
        rec_fields = [] # the name of the fields in the object or dictionary
        for (n1,n2) in results:
            if n2 is None:
                # "stand-alone" name
                if n1 in recdefs.keys():
                    sql_res += [ '%s.%s' %(n1,v) for v in self.names[n1] ]
                    rec_fields+=[ v for v in self.names[n1] ]
                else:
                    sql_res.append(n1)
                    rec_fields.append(n1)
            else:
                # "qualified" name, with the table name first
                sql_res.append('%s.%s' %(n1,n2))
                # in the result set, the object will have the attribute 
                # table_name (we can't set an attribute table.name, and
                # name alone could be ambiguous
                rec_fields.append('%s_%s' %(n1,n2))
        self.rec_fields = rec_fields
        
        # now we can build the actual SQL string
        sql = 'SELECT '+ ','.join(sql_res)
        sql += ' FROM '
        froms = []
        for (k,v) in recdefs.iteritems():
            froms.append('%s AS %s ' %(v,k))
        sql += ','.join(froms)
        if conditions:
            sql += 'WHERE '
        for c in conditions:
            sql += get_source(c)

        return sql
            
    def __iter__(self):
        return self
    
    def next(self):
        if not self._iterating:
            # begin iteration
            self.cursor.execute(self.sql)
            self._iterating = True
        row = self.cursor.fetchone()
        if row is not None:
            if self.return_type == object:
                # transform list into instance of Record
                # uses the rec_fields computed in build_sql()
                rec = Record()
                rec.__dict__ = dict(zip(self.rec_fields,row))
                return rec
            elif self.return_type == dict:
                return dict(zip(self.rec_fields,row))
            elif self.return_type == list:
                return row
        self._iterating = False
        raise StopIteration


if __name__ == "__main__":

    plane = Table()
    countries = Table()

    # all the items produced by iteration on query() are instances
    # of the Record class

    # simple requests
    # since no attribute of r is specified in the query, returns a list
    # of instances of Record with attributes matching all the field names
    q = query(r for r in plane if r.country == 'France').sql
    print(q)
    #
    # # this request returns a list instances of Record with the attribute
    # # c_country (c.country with the . replaced by _)
    # print[country
    # for country in query(c.country for c in countries
    #                      if c.continent == 'Europe') ]
    #
    # # request on two tables
    # print[r.name
    # for r in query(r for r in plane for c in countries
    #                if r.country == c.country and c.continent == 'Europe')]



from onto import domain_model, attrs
# from onto.collection_mixin import with_pony


class LocationBase(domain_model.DomainModel
    # , metaclass=with_pony
                   ):
    # doc_ref = attrs.doc_ref(type_cls=str)

    class Meta:
        collection_name = "locations"



class Location(LocationBase):

    @classmethod
    def _datastore(cls):
        from onto.database.kafka import KafkaDatabase
        return KafkaDatabase

    latitude = attrs.bproperty(type_cls=float)
    longitude = attrs.bproperty(type_cls=float)
    address = attrs.bproperty(type_cls=float)
    meetings = attrs.relation(dm_cls='Meeting', import_required=False, collection=list, nested=False)



from datetime import datetime
from onto import schema, fields, domain_model, utils, attrs
from onto.config import Config
from onto.models.mixin import Exportable, NewMixin
from onto.models.base import BaseRegisteredModel, Schemed
from onto.context import Context as CTX


if not CTX._ready:
    config = Config(
        app_name="flask-boiler-testing",
        debug=True,
        testing=True,
        certificate_filename="flask-boiler-testing-firebase-adminsdk-4m0ec-7505aaef8d.json"
    )

    CTX.read(config)

class Post(domain_model.DomainModel):

    _collection_name = "posts"

    _id = attrs.bproperty(data_key="id", import_enabled=False)

    @_id.getter
    def _id(self):
        return self.doc_id

    title = attrs.bproperty()
    body = attrs.bproperty()
    pub_date = attrs.bproperty(export_default=None)

    category = attrs.relation(dm_cls='Category', nested=True)

    def __repr__(self):
        return '<Post %r>' % self.title


class Category(domain_model.DomainModel):
    _id = attrs.bproperty(data_key="id", import_enabled=False)

    name = attrs.bproperty()
    posts = attrs.relation(dm_cls='Post', nested=True, collection=list)

    @_id.getter
    def _id(self):
        return self.doc_id

    _collection_name = "categories"

    def __repr__(self):
        return '<Category %r>' % self.name


category_id = utils.random_id()
py = Category.new(doc_id=category_id)
py.name = "Python"

post_id = utils.random_id()
p = Post.new(doc_id=post_id)
p.title = "snakes"
p.body = "Ssssssss"

# py.posts.append(p)
p.category = py

py.posts.append(p)

p.save()

obj = Post.get(doc_id=post_id)

assert str(obj.category) == "<Category 'Python'>"

print(obj)

# assert obj._export_as_view_dict() == {'body': 'Ssssssss',
#                                     'id': post_id,
#                                     'category': {
#                                         'id': category_id,
#                                         'name': 'Python'},
#                                     'title': 'snakes',
#                                     'pubDate': None}

assert p.category.doc_id == category_id



from onto import fields, schema, domain_model
from .. import domain_models
from tests.fixtures import CTX, setup_app
from .fixtures import users, tickets, location
from marshmallow import fields as m_fields


def test_reference(users, tickets, location):
    """ Used as an experiment
    """

    from onto import attrs

    class ExpMeeting(domain_model.DomainModel):

        class Meta:
            collection_name = "meetings"

        location = attrs.relation(nested=True, dm_cls=domain_models.Location)
        users = attrs.relation(nested=True, dm_cls=domain_models.User, collection=list)
        tickets = attrs.relation(nested=True, dm_cls=domain_models.Ticket, collection=dict)
        status = attrs.bproperty()

    m = ExpMeeting.new(doc_id="meeting_1")
    m.users = [user for user in users]
    m.tickets = {ticket.user.id: ticket for ticket in tickets}
    m.location = location
    m.status = "in-session"

    md = m.to_dict()

    mo = ExpMeeting.from_dict(d=md)
    u = mo.users[0]
    assert u is not None



import threading

# from google.cloud.firestore import DocumentReference

from .collection_mixin import CollectionMixin, CollectionMemberMeta
from .context import Context as CTX
from .models.base import Serializable
from .utils import random_id


class PersistableMixin:
    _struct_collection_id = "Persistable"

    def __init__(self, *args, struct_id=None, struct_d=None, **kwargs):
        super().__init__(*args, **kwargs)
        if struct_id is None:
            self._structure_id = random_id()
        else:
            self._structure_id = struct_id

        if struct_d is None:
            self._structure = dict()
        else:
            self._structure = struct_d
        # self._structure["vm_type"] = self.__class__.__name__

    def __get_ref(self) -> 'google.cloud.firestore.DocumentReference':
        return CTX.db.collection(self._struct_collection_id) \
            .document(self._structure_id)

    def _save__structure(self):
        self.__get_ref().set(self._structure)

    def _get__structure(self):
        return self.__get_ref().get()


class ViewModelMixin:
    """
    TODO: check if unsubscribe needs to be called on all on_update functions
            when deleting an instance.
    TODO: consider redundancy when ViewModel object becomes invalid in runtime
    TODO: consider decoupling bind_to to a superclass
    """

    @classmethod
    def get(cls, store=None, once=False, **kwargs):
        """ Returns an instance of view model with listener activated.

        :param struct_d: Binding structure.
        :param once: If set to True, do not listen to future document updates
        :return:
        """

        obj = cls.new(store=store, **kwargs)
        return obj

    @classmethod
    def get_many(cls, struct_d_iterable=None, once=False):
        """ Gets a list of view models from a list of
            binding structures.

        :param struct_d_iterable: Binding structure.
        :param once: If set to True, do not listen to future document updates
        :return:
        """
        return [cls.get(struct_d=struct_d, once=once)
                for struct_d in struct_d_iterable]

    def __init__(
            self, f_notify=None, store=None, *args, **kwargs):
        """

        :param f_notify: callback to notify that view model's
            business properties have finished updating
        :param args:
        :param kwargs:
        """
        self.store = store
        super().__init__(*args, **kwargs)
        self.f_notify = f_notify

        self.has_first_success = False
        self.success_condition = threading.Condition()
        self._notify()

    def propagate_change(self):
        """ Save all objects mutated in a mutation

        :return:
        """
        self.store.propagate_back()

    def wait_for_first_success(self):
        """ Blocks a thread until self.has_first_success is True.
        TODO: fix a bug where error may arise when a prior task crashes
            with an exception and the code here keeps waiting for
            self.has_first_success to become true and blocks the thread
            from executing the next task
        NOTE: the above-said behaviors happen when using self._notify
        :return:
        """
        with self.success_condition:
            while not self.has_first_success:
                self.success_condition.wait()

    def _notify(self):
        """ Notify that this object has been changed by underlying view models.
            Once this object has a different value for underlying domain models,
                save the object to Firestore. Note that this method is
                expected to be called only after the data is consistent.
                (Ex. When all relevant changes made in a single transaction
                    from another server has been loaded into the object.
                )
            NOTE: if f_notify is not set has_first_success will still
                be set to True, and wait_for_first_success() will return
                exactly after _notify is called (which is empty).
        """
        if self.f_notify is not None:
            self.f_notify(obj=self)
        with self.success_condition:
            self.success_condition.notify()
            self.has_first_success = True

    def to_view_dict(self):
        # with self.lock:
        return self._export_as_dict()

    # def _export_as_dict(self, *args, transaction=None, **kwargs):
    #     """ Must not modify datastore
    #
    #     :param args:
    #     :param transaction:
    #     :param kwargs:
    #     :return:
    #     """
    #     return self._export_as_view_dict(*args, **kwargs)

    def to_dict(self):
        return self.to_view_dict()


class ViewModel(ViewModelMixin, Serializable):
    """ View model are generated and refreshed automatically as domain
            model changes. Note that states stored in ViewModel
            are unreliable and should not be used to evaluate other states.
            Note that since ViewModel is designed to store in a database
            that is not strongly consistent, fields may be inconsistent.

    """
    pass
    # def __init__(self, *args, doc_ref=None, **kwargs):
    #     super().__init__(*args, doc_ref=doc_ref, **kwargs)


class ViewModelR(
    CollectionMixin, ViewModel, metaclass=CollectionMemberMeta):
    """
    ViewModelR = View Model with Repo
    """

    @classmethod
    def _datastore(cls):
        return CTX.db

    @classmethod
    def _get_collection(cls):
        return cls._datastore().ref / "**" / cls._get_collection_name()


import pytest


@pytest.mark.skip(reason='Context should be initialized with conftest already')
def test_firebase_app_context():
    from onto import config
    from onto import context

    Config = config.Config

    config = Config(
        app_name="flask-boiler-testing",
        debug=True,
        testing=True,
        certificate_filename="flask-boiler-testing-firebase-adminsdk-4m0ec-7505aaef8d.json"
    )
    CTX = context.Context
    CTX.read(config)
    assert CTX.firebase_app.project_id == "flask-boiler-testing"


@pytest.mark.skip(reason='Context should be initialized with conftest already')
def test_context_load():
    from onto.context import Context as CTX

    CTX.load()

    assert CTX.firebase_app.project_id == "flask-boiler-testing"


def test_config_comparator():
    from onto.config import Config
    config = Config.load()
    other_config = Config.load()
    assert config == other_config


def test_errors(monkeypatch):
    # Note: Watch out for state sharing
    # Note: Context may have invalid values when used in other test cases
    from onto.context import Context

    class CertFailError(Exception):
        pass

    def initialize_cert_fail(*args, **kwargs):
        raise CertFailError

    with monkeypatch.context() as m:
        import firebase_admin
        m.setattr(firebase_admin.credentials, "Certificate", initialize_cert_fail)
        with pytest.raises(CertFailError):
            Context._reload_credentials('./non-existent.json')

    # class AppFailError(Exception):
    #     pass
    #
    # def initialize_app_fail(*args, **kwargs):
    #     raise AppFailError
    #
    # with monkeypatch.context() as m:
    #     import firebase_admin
    #     m.setattr(firebase_admin, "initialize_app", initialize_app_fail)
    #     _PartialConfig = type(
    #         "_PartialConfig",
    #         (object,),
    #         {"APP_NAME": "AppNameFail"}
    #     )
    #     m.setattr(Context, "config", _PartialConfig())
    #     with pytest.raises(AppFailError):
    #         Context._reload_firebase_app()

    with pytest.raises(Exception):
        Context._reload_credentials(certificate_path="non-existent")



"""
Author: Zixuan Rao
Reference:
    http://flask.pocoo.org/docs/1.0/config/
    https://medium.freecodecamp.org/structuring-a-flask-restplus-web-service-for-production-builds-c2ec676de563
"""

import os

"""
Reads config from "boiler.yaml" in root directory if unspecified. 
Reads config from FB_CONFIG_FILENAME env var if specified.  
"""
CONFIG_FILENAME_KEY = "FB_CONFIG_FILENAME"
DEFAULT_CONFIG_FILENAME = "boiler.yaml"

caller_module_path = os.path.curdir
config_jsons_path = os.path.join(caller_module_path, "config_jsons")


class ConfigBase:
    """
    TESTING: (authenticate decorator will return uid as the value of header["Authentication"])
        Enable testing mode. Exceptions are propagated rather than handled by the the app’s error handlers.
        Extensions may also change their behavior to facilitate easier testing.
        You should enable this in your own tests.
    DEBUG:
        Whether debug mode is enabled.
        When using flask run to start the development server, an interactive debugger will be shown for unhandled
        exceptions, and the server will be reloaded when code changes. The debug attribute maps to this config key.
        This is enabled when ENV is 'development' and is overridden by the FLASK_DEBUG environment variable.
        It may not behave as expected if set in code.

    Note that this Config currently does not affect (Flask) main.app CONFIG.
    TODO: extend from Flask Config and apply to main.app

    """
    DEBUG: bool = None
    TESTING: bool = None
    FIREBASE_CERTIFICATE_JSON_PATH: str = None
    APP_NAME: str = None
    STORAGE_BUCKET_NAME: str = None

    def __eq__(self, other):
        """
        Comparator for configs to avoid reloading the same config.

        Note: this does not compare all fields

        :param other:
        :return:
        """
        if other is None:
            return False
        else:
            return self.APP_NAME == other.APP_NAME \
                and self.DEBUG == other.DEBUG \
                and self.TESTING == other.TESTING

    def __new__(cls, certificate_filename=None, certificate_path=None,
                testing=False, debug=False,
                app_name=None, storage_bucket_name=None,
                database=None, default_database=None, services=None,
                *args, **kwargs):
        cls.TESTING = testing
        cls.DEBUG = debug
        cls.APP_NAME = app_name

        cls.database = database if database is not None else dict()
        cls.default_database = default_database
        if services is None:
            services = dict()
        cls.services = services
        return cls


class Config(ConfigBase):

    @classmethod
    def _load_from_file(cls, file):
        from yaml import load, dump
        try:
            from yaml import CLoader as Loader, CDumper as Dumper
        except ImportError:
            from yaml import Loader, Dumper
        y = load(file, Loader=Loader)
        return cls(**y)

    @classmethod
    def load(cls):

        import os
        filename = os.getenv(CONFIG_FILENAME_KEY, DEFAULT_CONFIG_FILENAME)

        with open(filename) as file:
            return cls._load_from_file(file)



from google.cloud.firestore_v1 import Transaction

from onto.models.base import Serializable
from onto import schema, fields


class ReferencedObjectSchema(schema.Schema):

    doc_ref = fields.String(
        # attribute="doc_ref",
        # # dump_only=True,
        data_key="doc_ref",
        load_only=True,
        required=False,
        allow_none=True
    )


class ReferencedObject(Serializable):
    """
    ReferencedObject may be placed anywhere in the database.
    """
    _schema_base = ReferencedObjectSchema

    #
    # @property
    # def doc_ref(self):
    #     return self._doc_ref

    # @classmethod
    # def get(cls, *, doc_ref=None, transaction: Transaction = None, **kwargs):
    #     """ Returns an instance from firestore document reference.
    #
    #     :param doc_ref: firestore document reference
    #     :param transaction: firestore transaction
    #     :return:
    #     """
    #     return super().get(doc_ref=doc_ref, transaction=transaction)
